<?php
/**
 * 2007-2019 PrestaShop
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@prestashop.com so we can send you a copy immediately.
 *
 * DISCLAIMER
 *
 * Do not edit or add to this file if you wish to upgrade PrestaShop to newer
 * versions in the future. If you wish to customize PrestaShop for your
 * needs please refer to http://www.prestashop.com for more information.
 *
 * @author    Fattura24 <fattura24.com>
 * @copyright Copyright (c) Fattura24
 * @license   Commercial license
 *
 */

if (!defined('_PS_VERSION_')) {
    exit;
}

class Fattura24 extends Module
{
    public function __construct()
    {
        $this->baseUrl = 'https://www.app.fattura24.com/api/v0.3/';
        //$this->baseUrl = "https://www.fattura24.com/api/v0.3/";
        $this->name = 'fattura24';
        $this->tab = 'administration';
        $this->version = '2.5.1'; // modificare insieme al file config.xml
        $this->moduleDate = '2022/05/16 15:00';
        $this->author = 'Fattura24.com';
        $this->displayName = $this->l('Fattura24.com');
        $this->description = $this->l('Create your invoices by Fattura24.com');
        $this->module_key = '00cd0339f7de777834493205747ac875';
        $this->need_instance = 0;
        $this->bootstrap = true;
        $this->sourceInfo = 'F24-Pre ' . $this->version;
        $this->phpversion = phpversion();
        parent::__construct();
    }

    public function install()
    {
        $f24_fields = $this->f24Options();

        foreach ($f24_fields as $field) {
            if (strpos('MODELLO', $field) !== false) {
                $updatedValue = $this->l('Default');
            } elseif (strpos('SEZIONALE', $field) !== false) {
                $updatedValue = $this->l('Default');
            } elseif (strpos('SPEDIZIONE', $field) !== false) {
                $updatedValue = $this->l('None');
            } elseif (strpos('CREAZIONE', $field) !== false) {
                $updatedValue = $this->l('Disabled');
            } elseif (strpos('PDC', $field) !== false) {
                $updatedValue = $this->l('No Pdc');
            } elseif (strpos('BOLLO', $field) !== false) {
                $updatedValue = $this->l('Never');
            } elseif (strpos('MAIL', $field) !== false || strpos('PDF', $field) !== false) {
                $updatedValue = '0';
            } elseif (strpos('MAGAZZINO', $field) !== false) {
                $updatedValue = '0';
            } elseif (strpos('SALVA', $field) !== false || strpos('CREA', $field) !== false) {
                $updatedValue = '1';
            } else {
                $updatedValue = '';
            }

            if (Tools::strlen(Configuration::get($field)) == 0) {
                Configuration::updateValue($field, $updatedValue);
            }
            Configuration::updateValue('PS_F24_LOGFILE', $this->getLogFileName());
        }

        $dim = 'update `' . _DB_PREFIX_ . 'order_state` set pdf_invoice= \'';
        $dim .= pSQL('0') . '\' where id_order_state=\'2\'';
        Db::getInstance()->Execute($dim);

        /**
        * Gestione campi aggiuntivi con tabella
        */
        $sql = [];

        $sql[] = 'CREATE TABLE IF NOT EXISTS `' . _DB_PREFIX_ . 'fattura24` (
            `id_fattura24` int(11) NOT NULL AUTO_INCREMENT,
            `id_customer` int(11),
            `fattura24_codice_destinatario` varchar(13),
            `fattura24_pec` varchar(60),
            PRIMARY KEY  (`id_fattura24`)
        ) ENGINE=' . _MYSQL_ENGINE_ . ' DEFAULT CHARSET=utf8;';

        foreach ($sql as $query) {
            if (Db::getInstance()->execute($query) == false) {
                return false;
            }
        }

        if (!parent::install()) {
            return false;
        }

        /**
         * Nuovo metodo per aggiungere le nuove colonne alla tabella ordini di Prestashop
         * Le aggiungo solo se non esistono già. In caso di azione fallita scrivo nel log l'errore
         * Davide Iandoli 19.08.2020
         */
        $sql_orders = 'DESCRIBE `' . _DB_PREFIX_ . 'orders`';
        $fields = Db::getInstance(_PS_USE_SQL_SLAVE_)->ExecuteS($sql_orders);
        $new_fields = ['docIdFattura24', 'docIdOrderFattura24',
                            'apiResponseOrder', 'apiResponseInvoice', ];
        foreach ($new_fields as $new_field) {
            $existing_fields = false;
            foreach ($fields as $row) {
                if ($row['Field'] == $new_field) {
                    $existing_fields = true;
                } else {
                    if (!$existing_fields) {
                        try {
                            $query = 'ALTER TABLE `' . _DB_PREFIX_ . 'orders` ADD `' . $new_field .
                            '` varchar(500) DEFAULT NULL';
                            Db::getInstance()->execute($query);
                        } catch (PrestaShopDatabaseException $e) {
                            //$this->trace(' qualcosa è andato storto :', $e);
                        }
                    }
                }
            }
        }

        // registro gli hooks che utilizzo
        $fatt24Hooks = [
            'displayAdminCustomers',
            'displayAdminOrder',
            'displayCustomerAccountForm',
            'actionCustomerAccountAdd',
            'actionCustomerAccountUpdate',
            'actionValidateOrder',
            'actionOrderStatusUpdate',
            'actionOrderStatusPostUpdate'
        ];

        foreach ($fatt24Hooks as $hook) {
            if (!$this->registerHook($hook)) {
                $this->registerHook($hook);
            }
        }

        // crea cartella per il PDF
        $documentsFolder = _PS_DOWNLOAD_DIR_ . 'Fattura24/';
        if (!file_exists($documentsFolder)) {
            mkdir($documentsFolder, 0777, true);
        }

        if (!file_exists($documentsFolder . 'index.php')) {
            file_put_contents($documentsFolder . 'index.php', '<?php');
        }

        $this->trace('Modulo installato correttamente!');

        return true;
    }

    private function uninstallDB()
    {
        return Db::getInstance()->execute('
       DROP TABLE IF EXISTS ' . _DB_PREFIX_ . 'fattura24;');
    }

    public function uninstall()
    {
        if (!parent::uninstall()) {
            Configuration::deleteByName('PS_FATTURA24_API');
        }
        Configuration::deleteByName('PS_SALVA_CLIENTE');

        $this->uninstallDB();
        parent::uninstall();

        return true;
    }


    // ottiene il valore delle impostazioni
    public function getContent()
    {
        /**
         * Api calls to get lists. By this methos I call API only three times
         * and pass the result as a parameter to my method
         */
        $f24_template_url = $this->baseUrl . 'GetTemplate';
        $f24_pdc_url = $this->baseUrl . 'GetPdc';
        $f24_sezionali_url = $this->baseUrl . 'GetNumerator';
        $send_data = [];
        $send_data['apiKey'] = Configuration::get('PS_FATTURA24_API');
        $send_data['source'] = $this->sourceInfo;
        // fix introdotto per gestire i casi in cui il campo input API Key è vuoto
        if (!empty($send_data['apiKey'])) {
            $templates = $this->curlDownload($f24_template_url, http_build_query($send_data));
            $apiError = $this->getApiError($templates);
            // se è fallita la prima chiamata non faccio le altre due
            $f24_pdc = $apiError ? '' : $this->curlDownload(
                $f24_pdc_url,
                http_build_query($send_data)
            );
            $f24_sezionali = $apiError ? '' : $this->curlDownload(
                $f24_sezionali_url,
                http_build_query($send_data)
            );
        } else {
            $templates = '';
            $f24_pdc = '';
            $f24_sezionali = '';
        }

        $output = $this->getVersionCheckMessage() . $this->getNaturaError()
                    . $this->getApiError($templates);

        if (Tools::isSubmit('submitModule')) {
            Configuration::updateValue('PS_FATTURA24_API', $_REQUEST['PS_FATTURA24_API']);

            if (isset($_REQUEST['PS_CONF_NATURA_IVA'])) {
                Configuration::updateValue('PS_CONF_NATURA_IVA', $_REQUEST['PS_CONF_NATURA_IVA']);
            }
            if (isset($_REQUEST['PS_CONF_NATURA_SPEDIZIONE'])) {
                Configuration::updateValue('PS_CONF_NATURA_SPEDIZIONE', $_REQUEST['PS_CONF_NATURA_SPEDIZIONE']);
            }
            if (isset($_REQUEST['PS_INV_OBJECT'])) {
                Configuration::updateValue('PS_INV_OBJECT', $_REQUEST['PS_INV_OBJECT']); // oggetto del documento
            }
            //se le checkbox sono settate imposto il valore della relativa configurazione a 1, altrimenti a 0
            if (isset($_REQUEST['PS_SALVA_CLIENTE_'])) {
                Configuration::updateValue('PS_SALVA_CLIENTE', $_REQUEST['PS_SALVA_CLIENTE_']);
            } else {
                Configuration::updateValue('PS_SALVA_CLIENTE', '0');
            }

            if (isset($_REQUEST['PS_CREA_ORDINE_'])) {
                Configuration::updateValue('PS_CREA_ORDINE', $_REQUEST['PS_CREA_ORDINE_']);
            } else {
                Configuration::updateValue('PS_CREA_ORDINE', '0');
            }

            if (isset($_REQUEST['PS_ABILITA_ORDINE_ZERO'])) {
                Configuration::updateValue('PS_ABILITA_ORDINE_ZERO', $_REQUEST['PS_ABILITA_ORDINE_ZERO']);
            } else {
                Configuration::updateValue('PS_ABILITA_ORDINE_ZERO', '0');
            }

            if (isset($_REQUEST['PS_NUMERO_ORDINE'])) {
                Configuration::updateValue('PS_NUMERO_ORDINE', $_REQUEST['PS_NUMERO_ORDINE']);
            } else {
                Configuration::updateValue('PS_NUMERO_ORDINE', '0');
            }

            if (isset($_REQUEST['PS_PDF_ORDINE_'])) {
                Configuration::updateValue('PS_PDF_ORDINE', $_REQUEST['PS_PDF_ORDINE_']);
            } else {
                Configuration::updateValue('PS_PDF_ORDINE', '0');
            }

            if (isset($_REQUEST['PS_MAIL_ORDINE_'])) {
                Configuration::updateValue('PS_MAIL_ORDINE', $_REQUEST['PS_MAIL_ORDINE_']);
            } else {
                Configuration::updateValue('PS_MAIL_ORDINE', '0');
            }

            if (isset($_REQUEST['PS_F24_MODELLO_ORDINE_'])) {
                Configuration::updateValue('PS_F24_MODELLO_ORDINE', $_REQUEST['PS_F24_MODELLO_ORDINE_']);
            } else {
                Configuration::updateValue('PS_F24_MODELLO_ORDINE', $this->l('Default'));
            }

            if (isset($_REQUEST['PS_F24_MODELLO_ORDINE_DEST_'])) {
                Configuration::updateValue('PS_F24_MODELLO_ORDINE_DEST', $_REQUEST['PS_F24_MODELLO_ORDINE_DEST_']);
            } else {
                Configuration::updateValue('PS_F24_MODELLO_ORDINE_DEST', $this->l('Default'));
            }

            if (isset($_REQUEST['PS_CREAZIONE_FATTURA_'])) {
                Configuration::updateValue('PS_CREAZIONE_FATTURA', $_REQUEST['PS_CREAZIONE_FATTURA_']);
            } else {
                Configuration::updateValue('PS_CREAZIONE_FATTURA', '0');
            }

            if (isset($_REQUEST['PS_ABILITA_FATTURA_ZERO'])) {
                Configuration::updateValue('PS_ABILITA_FATTURA_ZERO', $_REQUEST['PS_ABILITA_FATTURA_ZERO']);
            } else {
                Configuration::updateValue('PS_ABILITA_FATTURA_ZERO', '0');
            }

            if (isset($_REQUEST['PS_EMAIL_FATTURA_'])) {
                Configuration::updateValue('PS_EMAIL_FATTURA', $_REQUEST['PS_EMAIL_FATTURA_']);
            } else {
                Configuration::updateValue('PS_EMAIL_FATTURA', '0');
            }

            if (isset($_REQUEST['PS_STATO_PAGATO_'])) {
                Configuration::updateValue('PS_STATO_PAGATO', $_REQUEST['PS_STATO_PAGATO_']);
            } else {
                Configuration::updateValue('PS_STATO_PAGATO', '0');
            }

            if (isset($_REQUEST['PS_DISABILITA_RICEVUTE_'])) {
                Configuration::updateValue('PS_DISABILITA_RICEVUTE', $_REQUEST['PS_DISABILITA_RICEVUTE_']);
            } else {
                Configuration::updateValue('PS_DISABILITA_RICEVUTE', '0');
            }

            if (isset($_REQUEST['PS_F24_MODELLO_FATTURA_'])) {
                Configuration::updateValue('PS_F24_MODELLO_FATTURA', $_REQUEST['PS_F24_MODELLO_FATTURA_']);
            } else {
                Configuration::updateValue('PS_F24_MODELLO_FATTURA', $this->l('Default'));
            }

            if (isset($_REQUEST['PS_F24_MODELLO_FATTURA_DEST_'])) {
                Configuration::updateValue('PS_F24_MODELLO_FATTURA_DEST', $_REQUEST['PS_F24_MODELLO_FATTURA_DEST_']);
            } else {
                Configuration::updateValue('PS_F24_MODELLO_FATTURA_DEST', $this->l('Default'));
            }

            if (isset($_REQUEST['PS_F24_PDC_'])) {
                Configuration::updateValue('PS_F24_PDC', $_REQUEST['PS_F24_PDC_']);
            } else {
                Configuration::updateValue('PS_F24_PDC', $this->l('Default'));
            }

            if (isset($_REQUEST['PS_F24_SEZIONALE_RICEVUTA_'])) {
                Configuration::updateValue('PS_F24_SEZIONALE_RICEVUTA', $_REQUEST['PS_F24_SEZIONALE_RICEVUTA_']);
            } else {
                Configuration::updateValue('PS_F24_SEZIONALE_RICEVUTA', $this->l('Default'));
            }

            if (isset($_REQUEST['PS_F24_SEZIONALE_FATTURA_'])) {
                Configuration::updateValue('PS_F24_SEZIONALE_FATTURA', $_REQUEST['PS_F24_SEZIONALE_FATTURA_']);
            } else {
                Configuration::updateValue('PS_F24_SEZIONALE_FATTURA', $this->l('Default'));
            }
            if (isset($_REQUEST['PS_F24_BOLLO_VIRTUALE_FE'])) {
                Configuration::updateValue('PS_F24_BOLLO_VIRTUALE_FE', $_REQUEST['PS_F24_BOLLO_VIRTUALE_FE']);
            }

            $output = $this->getVersionCheckMessage() . $this->getNaturaError() . $this->getApiError($templates);
            $output .= $this->displayConfirmation($this->l('Updated settings'));
        }
        // i valori delle impostazioni vengono assegnati a smarty, che li userà per i template (file *.tpl)
        $this->context->smarty->assign(
            [
                'token' => Tools::getValue('token'), // fix per PS 1.7.7
                'isFattura24Enabled' => Module::isEnabled('fattura24'),
                'psDisableOverrides' => Configuration::get('PS_DISABLE_OVERRIDES'),
                'version' => $this->version,
                'source' => $this->sourceInfo,
                'apiKey' => Configuration::get('PS_FATTURA24_API'),
                'listaModelliOrdine' => $this->getTemplate($templates, true),
                'numOrdine' => $this->f24OrderNumber(),
                'numOrdineSelected' => Configuration::get('PS_NUMERO_ORDINE'),
                'modelloOrdine' => Configuration::get('PS_F24_MODELLO_ORDINE'),
                'modelloOrdineDest' => Configuration::get('PS_F24_MODELLO_ORDINE_DEST'),
                'listaModelliFattura' => $this->getTemplate($templates, false),
                'opzioniCreaFattura' => $this->getInvoiceOptions(),
                'tipiAliquoteIva' => $this->getListaNatureNew(),
                'opzioniStatoPagato' => $this->f24PaidStatus(),
                'modelloFattura' => Configuration::get('PS_F24_MODELLO_FATTURA'),
                'modelloFatturaDest' => Configuration::get('PS_F24_MODELLO_FATTURA_DEST'),
                'listaConti' => $this->getPdc($f24_pdc),
                'conto' => Configuration::get('PS_F24_PDC'),
                'salvaCliente' => Configuration::get('PS_SALVA_CLIENTE'),
                'creazioneOrdine' => Configuration::get('PS_CREA_ORDINE'),
                'abilitaOrdineZero' => Configuration::get('PS_ABILITA_ORDINE_ZERO'),
                'pdfOrdine' => Configuration::get('PS_PDF_ORDINE'),
                'emailOrdine' => Configuration::get('PS_MAIL_ORDINE'),
                'creazioneFattura' => Configuration::get('PS_CREAZIONE_FATTURA'),
                'abilitaFatturaZero' => Configuration::get('PS_ABILITA_FATTURA_ZERO'),
                'confNaturaIva' => Configuration::get('PS_CONF_NATURA_IVA'),
                'naturaSpedizione' => Configuration::get('PS_CONF_NATURA_SPEDIZIONE'),
                'invObject' => Configuration::get('PS_INV_OBJECT'), // oggetto del documento
                'emailFattura' => Configuration::get('PS_EMAIL_FATTURA'),
                'statoPagato' => Configuration::get('PS_STATO_PAGATO'),
                'disabilitaRicevute' => Configuration::get('PS_DISABILITA_RICEVUTE'),
                'listaSezionaliFattura' => $this->getNumerator($f24_sezionali, 1),
                'listaSezionaliFatturaElettronica' => $this->getNumerator($f24_sezionali, 11),
                'listaSezionaliRicevuta' => $this->getNumerator($f24_sezionali, 3),
                'sezionaleRicevuta' => Configuration::get('PS_F24_SEZIONALE_RICEVUTA'),
                'sezionaleFattura' => Configuration::get('PS_F24_SEZIONALE_FATTURA'),
                'bolloVirtualeFe' => $this->f24StampOptions(),
                'bollo' => Configuration::get('PS_F24_BOLLO_VIRTUALE_FE'),
                'moduleUrl' => _MODULE_DIR_ . 'fattura24/',
                'logFileName' => $this->getLogFileName(),
                'info' => $this->getInfo(),
            ]
        );

        $output .= $this->display(__FILE__, 'views/templates/admin/form.tpl'); // tutti i dati passano al file form.tpl

        return $output;
    }


    // valori immessi nella schermata impostazioni
    public function getConfigFieldsValues()
    {
        return [
            $this->f24Value('PS_FATTURA24_API'),
            $this->f24Value('PS_SALVA_CLIENTE'),
            $this->f24Value('PS_CREA_ORDINE_'),
            $this->f24Value('PS_ABILITA_ORDINE_ZERO'),
            $this->f24Value('PS_PDF_ORDINE_'),
            $this->f24Value('PS_MAIL_ORDINE_'),
            $this->f24Value('PS_F24_MODELLO_ORDINE_'),
            $this->f24Value('PS_F24_MODELLO_ORDINE_DEST_'),
            $this->f24Value('PS_CREAZIONE_FATTURA_'),
            $this->f24Value('PS_ABILITA_FATTURA_ZERO'),
            $this->f24Value('PS_CONF_NATURA_IVA'),
            $this->f24Value('PS_CONF_NATURA_SPEDIZIONE'),
            $this->f24Value('PS_INV_OBJECT'),
            $this->f24Value('PS_EMAIL_FATTURA_'),
            $this->f24Value('PS_STATO_PAGATO_'),
            $this->f24Value('PS_DISABILITA_RICEVUTE_'),
            $this->f24Value('PS_F24_MAGAZZINO_ORDINE_'),
            $this->f24Value('PS_F24_MAGAZZINO_FATTURA_'),
            $this->f24Value('PS_F24_MODELLO_FATTURA_'),
            $this->f24Value('PS_F24_MODELLO_FATTURA_DEST_'),
            $this->f24Value('PS_F24_PDC_'),
            $this->f24Value('PS_F24_SEZIONALE_FATTURA_'),
            $this->f24Value('PS_F24_SEZIONALE_RICEVUTA_'),
            $this->f24Value('PS_F24_BOLLO_VIRTUALE_FE'),
        ];
    }

    public function hookDisplayAdminCustomers($params)
    {
        if (Tools::substr(_PS_VERSION_, 0, 3) >= 1.7) {
            $idc = $params['id_customer'];
            $template_path = 'views/templates/admin/admin_customer.tpl'; //template per ps 1.7
        } else {
            $idc = pSQL(Tools::getValue('id_customer'));
            $template_path = 'views/templates/admin/admin_customer16.tpl'; //template per ps 1.6
        }

        $sdi_code = Tools::getValue('fattura24_codice_destinatario');
        if (!$sdi_code or Tools::strlen($sdi_code) <= 1) {
            $sdi_code = '0000000';
        }

        if (Tools::getValue('id_fattura24') > 0) {
            $sql = 'update `' . _DB_PREFIX_ . "fattura24` set fattura24_codice_destinatario = '";
            $sql .= pSQL((string) $sdi_code) . "',fattura24_pec='";
            $sql .= pSQL(Tools::getValue('fattura24_pec')) . "' where id_fattura24 = ";
            $sql .= (int) Tools::getValue('id_fattura24') . ' ';
            Db::getInstance()->Execute($sql);
        } else {
            $sql = 'INSERT INTO `' . _DB_PREFIX_ . 'fattura24`(id_customer,fattura24_codice_destinatario,fattura24_pec) 
                    VALUES (' . (int) $idc . ",'" . pSQL((string) $sdi_code) . "','"
                    . pSQL(Tools::getValue('fattura24_pec')) . "')";
            Db::getInstance()->Execute($sql);
        }

        $sql = 'select id_fattura24,fattura24_codice_destinatario,fattura24_pec from `';
        $sql .= _DB_PREFIX_ . 'fattura24` where id_customer = ' . (int) $idc . '';
        $result = Db::getInstance()->getRow($sql);

        if (Tools::substr(_PS_VERSION_, 0, 5) >= 1.7) {
            $linkController = Context::getContext()->link->getAdminLink('AdminCustomers', true, [], [
                'id_customer' => $idc,
                'viewcustomer' => 1,
            ]);
        } else {
            $linkController = $this->context->link->getAdminLink('AdminCustomers', true);
            $linkController .= '&id_customer=' . $idc . '&viewcustomer';
        }

        $this->context->smarty->assign('linkController', $linkController);
        $this->context->smarty->assign('id_fattura24', $result['id_fattura24']);
        $this->context->smarty->assign('fattura24_codice_destinatario', $result['fattura24_codice_destinatario']);
        $this->context->smarty->assign('fattura24_pec', $result['fattura24_pec']);

        $output = $this->context->smarty->fetch($this->local_path . $template_path);

        return $output;
    }

    /**
     * Nuovo hook aggiunti per controllo stato ordini/fatture in F24
     * Davide Iandoli 27.09.2019 - modificato il 27.02.2020
     */
    public function hookDisplayAdminOrder(array $params)
    {
        $orderId = (int) $params['id_order'];
        $order = new Order($orderId);
        $order_status = $order->getCurrentOrderState();
        if (is_object($order_status)) {
            $status_name = current($order_status->name);
            if ($status_name == $this->l('Cancelled')) {
                $order_status_name = '';
            } else {
                $order_status_name = $status_name;
            }
        } else {
            $order_status_name = $order_status;
        }
        $query = 'SELECT docIdFattura24, docIdOrderFattura24 FROM ' . _DB_PREFIX_;
        $query .= 'orders WHERE id_order=\'' . (int) $orderId . '\';';
        $docIdFattura24 = Db::getInstance(_PS_USE_SQL_SLAVE_)->getRow($query)['docIdFattura24'];
        $docIdOrderFattura24 = Db::getInstance(_PS_USE_SQL_SLAVE_)->getRow($query)['docIdOrderFattura24'];
        $new_query = '';

        /*
         * Con questa try/catch gestisco i casi in cui le colonne apiResponseOrder e apiResponseInvoice
         * non esistono. Fix del 19.08.2020
         */

        try {
            $new_query = 'SELECT apiResponseOrder, apiResponseInvoice from ' . _DB_PREFIX_ .
            $new_query .= 'orders WHERE id_order=\'' . (int) $orderId . '\';';
            $apiResponseOrder = Db::getInstance(_PS_USE_SQL_SLAVE_)->getRow($new_query)['apiResponseOrder'];
            $apiResponseInvoice = Db::getInstance(_PS_USE_SQL_SLAVE_)->getRow($new_query)['apiResponseInvoice'];
            $obj_apiResponseOrder = json_decode($apiResponseOrder);
            $obj_apiResponseInvoice = json_decode($apiResponseInvoice);
            if (is_object($obj_apiResponseOrder)) {
                $orderResponseDescription = $obj_apiResponseOrder->description;
            } else {
                $orderResponseDescription = '';
            }

            if (is_object($obj_apiResponseInvoice)) {
                $invoiceResponseDescription = $obj_apiResponseInvoice->description;
            } else {
                $invoiceResponseDescription = '';
            }
        } catch (PrestaShopDatabaseException $e) {
            //se le colonne non esistono
            //$this->trace(' errore nel db : ', $e);
            $orderResponseDescription = '';
            $invoiceResponseDescription = '';
        }

        $config_CreaOrdine = Configuration::get('PS_CREA_ORDINE');
        $config_CreaFattura = Configuration::get('PS_CREAZIONE_FATTURA');
        $config_Overrides = Configuration::get('PS_DISABLE_OVERRIDES');

        $order_documents = 0;
        if (!empty($docIdOrderFattura24 || strpos($orderResponseDescription, 'already exists') !== false)) {
            ++$order_documents;
        }

        if (!empty($docIdFattura24 || strpos($invoiceResponseDescription, 'already exists') !== false)) {
            ++$order_documents;
        }

        if (!$config_CreaOrdine) {
            $message_order = $this->l('invoice creation disabled');
        } elseif (empty($order_status_name)) {
            $message_order = $this->l('order cancelled in Prestashop');
        } else {
            if (!empty($docIdOrderFattura24 || strpos($orderResponseDescription, 'already exists') !== false)) {
                $message_order = $this->l('document saved in Fattura24');
            } elseif (strpos($orderResponseDescription, 'Invalid ID Request') !== false) {
                $message_order = $this->l('processing data in Fattura24, retry later');
            } else {
                $message_order = $this->l('document NOT saved in Fattura24');
            }
        }

        if (!$config_CreaFattura) {
            $message_invoice = $this->l('document creation disabled');
        } elseif (empty($order_status_name)) {
            $message_invoice = $this->l('order cancelled in Prestashop');
        } else {
            if (!empty($docIdFattura24 || strpos($invoiceResponseDescription, 'already exists') !== false)) {
                if (isset($obj_apiResponseInvoice->docNumber)) {
                    try {
                        $message_invoice = $this->l('doc n. ')
                        . $obj_apiResponseInvoice->docNumber
                        . $this->l(' saved in Fattura24');
                    } catch (Exception $e) {
                        $message_invoice = $this->l('document saved in Fattura24');
                    }
                } else {
                    $message_invoice = $this->l('document saved in Fattura24');
                }
            } elseif (strpos($invoiceResponseDescription, 'Invalid ID Request') !== false) {
                $message_invoice = $this->l('processing data in Fattura24, retry later');
            } else {
                $message_invoice = $this->l('document NOT saved in Fattura24');
            }
        }

        // uso order_documents per incrementare il contatore

        $this->context->smarty->assign('id_order', $orderId);
        $this->context->smarty->assign('order_status_name', $order_status_name);
        $this->context->smarty->assign('config_CreaOrdine', $config_CreaOrdine);
        $this->context->smarty->assign('config_CreaFattura', $config_CreaFattura);
        $this->context->smarty->assign('config_Overrides', $config_Overrides);
        $this->context->smarty->assign('order_documents', $order_documents); // contatore documenti salvati
        $this->context->smarty->assign('fattura24_invoice_docId', $docIdFattura24); //docId per tasto fattura
        $this->context->smarty->assign('fattura24_invoice', $message_invoice);
        $this->context->smarty->assign('fattura24_order_docId', $docIdOrderFattura24); //docId per tasto ordine
        $this->context->smarty->assign('fattura24_order', $message_order);
        if ((Tools::substr(_PS_VERSION_, 0, 5) >= '1.7.7')) {
            return $this->display(__FILE__, 'views/templates/hook/displayAdminOrderMainBottom.tpl');
        } else {
            return $this->display(__FILE__, 'views/templates/hook/displayAdminOrderLeft.tpl');
        }
    }

    /**
     * Hooks per la gestione dei campi SDI e PEC
     *  Davide Iandoli 08.03.2019
     */
    public function hookDisplayCustomerAccountForm($params)
    {
        if ($params['cookie']->id_customer) {
            $sql = 'select id_fattura24,fattura24_codice_destinatario,fattura24_pec from `' . _DB_PREFIX_;
            $sql .= 'fattura24` where id_customer = ' . (int) $params['cookie']->id_customer . '';
            $result = Db::getInstance()->getRow($sql);
            if ($result) {
                $this->context->smarty->assign('id_fattura24', $result['id_fattura24']);
                $this->context->smarty->assign('fattura24_id_customer', (int) $params['cookie']->id_customer);
                $this->context->smarty->assign(
                    'fattura24_codice_destinatario',
                    $result['fattura24_codice_destinatario']
                );
                $this->context->smarty->assign('fattura24_pec', $result['fattura24_pec']);
            } else {
                $this->context->smarty->assign('id_fattura24', '');
                $this->context->smarty->assign('fattura24_id_customer', 0);
                $this->context->smarty->assign('fattura24_codice_destinatario', '');
                $this->context->smarty->assign('fattura24_pec', '');
            }

            return $this->display(__FILE__, 'views/templates/hook/customer_reg_form.tpl');

        /* gestione account in caso di checkout ospite corretto il 21.03.2019 - fix per ps 1.7 - */
        } else {
            $this->context->smarty->assign('id_fattura24', '');
            $this->context->smarty->assign('fattura24_id_customer', (int) $params['cookie']->id_customer);
            $this->context->smarty->assign('fattura24_codice_destinatario', '');
            $this->context->smarty->assign('fattura24_pec', '');

            return $this->display(__FILE__, 'views/templates/hook/customer_reg_form.tpl');
        }
    }

    public function hookDisplayCustomerIdentityForm($params)
    {
        $sdi_code = Tools::getValue('fattura24_codice_destinatario');
        if (!$sdi_code or Tools::strlen($sdi_code) <= 1) {
            $sdi_code = '0000000';
        }

        if (Tools::getValue('id_fattura24') > 0) {
            $sql = 'update `' . _DB_PREFIX_ . "fattura24` set fattura24_codice_destinatario = '";
            $sql .= pSQL((string) $sdi_code) . "',fattura24_pec='" . pSQL((int) Tools::getValue('fattura24_pec'));
            $sql .= "' where id_fattura24 = " . (int) Tools::getValue('id_fattura24') . ' ';
            Db::getInstance()->Execute($sql);
        } elseif (Tools::getValue('fattura24_id_customer')) {
            $sql = 'INSERT INTO `' . _DB_PREFIX_ . 'fattura24`(id_customer,fattura24_codice_destinatario,fattura24_pec) 
                    VALUES (' . pSQL((int) Tools::getValue('fattura24_id_customer')) . ",'"
                    . pSQL((string) $sdi_code) . "','" . pSQL((string) Tools::getValue('fattura24_pec')) . "')";
            Db::getInstance()->Execute($sql);
        }

        if ($params['cookie']->id_customer) {
            $sql = 'select id_fattura24,fattura24_codice_destinatario,fattura24_pec from `' . _DB_PREFIX_;
            $sql .= 'fattura24` where id_customer = ' . (int) $params['cookie']->id_customer . '';
            $result = Db::getInstance()->getRow($sql);
            $customer = $result['id_customer'];

        /* gestione account ospite fix per ps 1.7 */
        } elseif ($params['cookie']->id_guest) {
            $sql = 'select id_fattura24,fattura24_codice_destinatario,fattura24_pec from `' . _DB_PREFIX_;
            $sql .= 'fattura24` where id_customer = ' . (int) $params['cookie']->id_guest . '';
            $result = Db::getInstance()->getRow($sql);
            $customer = $result['id_guest'];
        }

        $this->context->smarty->assign('id_fattura24', $result['id_fattura24']);
        $this->context->smarty->assign('fattura24_id_customer', $customer);
        $this->context->smarty->assign('fattura24_codice_destinatario', $result['fattura24_codice_destinatario']);
        $this->context->smarty->assign('fattura24_pec', $result['fattura24_pec']);

        return $this->display(__FILE__, 'views/templates/hook/customer_reg_form.tpl');
    }

    public function hookActionCustomerAccountAdd($params)
    {
        $newCustomer = $params['newCustomer'];
        $sql = 'INSERT INTO `' . _DB_PREFIX_ . 'fattura24`(id_customer,fattura24_codice_destinatario,fattura24_pec)
                  VALUES (' . (int) $newCustomer->id . ",'" . pSQL((string)
                  Tools::getValue('fattura24_codice_destinatario'))
                  . "','" . pSQL((string) Tools::getValue('fattura24_pec')) . "')";
        Db::getInstance()->Execute($sql);

        if (0) {
            $handle = fopen(dirname(__FILE__) . '/notify.html', 'a');
            fwrite($handle, "*********new*********<br>\n");
            fwrite($handle, 'start: ' . " \n<br>\n [" . date('Y-m-d h-m-s') . "]<br><br>\n");

            fwrite($handle, 'pec:' . Tools::getValue('fattura24_pec'));
            fwrite($handle, 'codice_destinatario:' . Tools::getValue('fattura24_codice_destinatario'));
            fwrite($handle, 'customer:' . $newCustomer->id);
            fwrite($handle, 'sql:' . $sql);

            fwrite($handle, 'end: ' . " \n<br>\n [" . date('Y-m-d h-m-s') . "]<br><br>\n");
            fclose($handle);
        }
    }

    //aggiorna i campi fiscali nell'account utente
    public function hookActionCustomerAccountUpdate($params)
    {
        $sdi_code = Tools::getValue('fattura24_codice_destinatario');
        if (!$sdi_code or Tools::strlen($sdi_code) <= 1) {
            $sdi_code = '0000000';
        }

        if (Tools::getValue('id_fattura24') > 0) {
            $sql = 'update `' . _DB_PREFIX_ . "fattura24` set fattura24_codice_destinatario = '";
            $sql .= pSQL((string) $sdi_code) . "',fattura24_pec='" . pSQL((string) Tools::getValue('fattura24_pec'));
            $sql .= "' where id_fattura24 = " . (int) Tools::getValue('id_fattura24') . ' ';
            Db::getInstance()->Execute($sql);
        } elseif (Tools::getValue('fattura24_id_customer')) {
            $sql = 'INSERT INTO `' . _DB_PREFIX_ . 'fattura24`(id_customer,fattura24_codice_destinatario,fattura24_pec) 
                    VALUES (' . pSQL((int) Tools::getValue('fattura24_id_customer')) . ",'"
                    . pSQL((string) $sdi_code) . "','" . pSQL((string) Tools::getValue('fattura24_pec')) . "')";
            Db::getInstance()->Execute($sql);
        }
    }

    /*fine hook aggiuntivi */

    public function hookActionValidateOrder($order)
    {
        if (!$this->active) {
            return;
        }

        $check_order = isset($order['order']) ? $order['order'] : $order['objOrder'];
        $zeroTotal = (float) $check_order->total_paid == 0.00;
        $zeroOrderEnabled = (int) Configuration::get('PS_ABILITA_ORDINE_ZERO') == 1;
        $sendOrder = (int) Configuration::get('PS_CREA_ORDINE') == 1;
        $sendOrder = $zeroTotal && !$zeroOrderEnabled ? false : true;

        if ($sendOrder) {
            $this->afterHook($check_order, false);
        } else {
            $this->saveCustomer($check_order);
        }
    }

    public function hookActionOrderStatusUpdate($order)
    {
        $orderPaid = (int) $order['newOrderStatus']->paid == 1;
        //$this->trace('order :', $order);
        //$this->trace('order paid 1 :', $orderPaid);
        if (!$this->active || !$orderPaid || (int) Configuration::get('PS_CREAZIONE_FATTURA') == 0) {
            return;
        }
        $orderId = $order['id_order'];
        $check_order = new Order($orderId);
        $zeroTotal = (float) $check_order->total_paid == 0.00;
        $zeroInvoiceEnabled = (int) Configuration::get('PS_ABILITA_FATTURA_ZERO') == 1;

        $query = 'SELECT docIdFattura24 FROM ' . _DB_PREFIX_ . 'orders WHERE id_order=\'' . (int) $orderId . '\';';
        $docIdFattura24 = Db::getInstance(_PS_USE_SQL_SLAVE_)->getRow($query)['docIdFattura24'];
        $sendInvoice = $zeroTotal && !$zeroInvoiceEnabled ? false : true;

        if ($sendInvoice && !$docIdFattura24) {
            $this->afterHook($check_order, true);
        }
    }

    // hook per rinominare il file pdf
    public function hookActionOrderStatusPostUpdate($order)
    {
        $orderPaid = (int) $order['newOrderStatus']->paid == 1;
        //$this->trace('order 2 :', $order);
        //$this->trace('order paid 2 :', $orderPaid);
        if (!$this->active || !$orderPaid|| (int) Configuration::get('PS_CREAZIONE_FATTURA') == 0) {
            return;
        }
        $orderId = $order['id_order'];
        $check_order = new Order($orderId);
        $number = $check_order->invoice_number;
        if (empty($number)) {
            return;
        }

        $newFileName = $this->constructFileName($number, $check_order->date_upd);
        $year = date('Y', strtotime($check_order->date_upd));
        $oldFileName = 'FA000000-' . $year . '.pdf';

        if (file_exists($oldFileName)) {
            //Qui viene fatta la rename del pdf
            $folder = _PS_DOWNLOAD_DIR_ . 'Fattura24/';
            rename($folder . $oldFileName, $folder . $newFileName);
        }
    }

    public function constructFileName($invoice_number, $date)
    {
        $year = date('Y', strtotime($date));

        return 'FA' . sprintf('%06d', $invoice_number) . '-' . $year . '.pdf';
    }

    // crea il contatto in xml
    public function createCustomerXml($order_id)
    {
        $this->trace('Create Customer Xml');
        $domtree = new DOMDocument('1.0', 'UTF-8');
        $domtree->preserveWhiteSpace = false;
        $domtree->formatOutput = true;
        $xmlRoot = $domtree->createElement('Fattura24');
        $xmlRoot = $domtree->appendChild($xmlRoot);
        $xmlCustomer = $domtree->createElement('Document');
        $xmlCustomer = $xmlRoot->appendChild($xmlCustomer);

        $invoice_address = new Order($order_id);
        $invoice_address = $invoice_address->id_address_invoice;

        $billing_data = new Address($invoice_address);
        $customer = new Customer($billing_data->id_customer);

        /**
         * Recupera i campi SDI e PEC per fatturazione elettronica
         * Davide Iandoli 08.03.2019
         */
        $sql = 'select id_fattura24,fattura24_codice_destinatario,fattura24_pec from `' . _DB_PREFIX_;
        $sql .= 'fattura24` where id_customer = ' . (int) $billing_data->id_customer . '';
        $result = Db::getInstance()->getRow($sql);
        $pec = $result['fattura24_pec'];
        $sdi_code = $result['fattura24_codice_destinatario'];

        /* controllo su codice destinatario */
        $isoCountry = Country::getIsoById($billing_data->id_country);

        if ($isoCountry != 'IT') {
            $sdi_code = 'XXXXXXX';
        } elseif (empty($result['fattura24_codice_destinatario'])) {
            $sdi_code = '0000000';
        } else {
            $sdi_code = $result['fattura24_codice_destinatario'];
        }

        $customer_fePec = $domtree->createElement('FeCustomerPec');
        $customer_fePec->appendChild($domtree->createCDataSection($pec));
        $xmlCustomer->appendChild($customer_fePec);

        $customer_feCodiceDestinatario = $domtree->createElement('FeDestinationCode');
        $customer_feCodiceDestinatario->appendChild($domtree->createCDataSection($sdi_code));
        $xmlCustomer->appendChild($customer_feCodiceDestinatario);

        $company = empty($customer->company) ? trim($billing_data->company) : trim($customer->company);

        if (!empty($company)) {
            $customername = $company;
        } else {
            $customername = $billing_data->firstname . ' ' . $billing_data->lastname;
        }

        $customer_name = $domtree->createElement('CustomerName');
        $customer_name->appendChild($domtree->createCDataSection($customername));
        $xmlCustomer->appendChild($customer_name);

        $customer_address = $domtree->createElement('CustomerAddress');
        $customer_address2 = $billing_data->address2 == null ? '' : ' ' . $billing_data->address2;
        $customer_address->appendChild($domtree->createCDataSection($billing_data->address1 . $customer_address2));

        $xmlCustomer->appendChild($customer_address);
        $xmlCustomer->appendChild($domtree->createElement('CustomerPostcode', $billing_data->postcode));
        $xmlCustomer->appendChild($domtree->createElement('CustomerCity', $billing_data->city));
        if (!empty($billing_data->id_state)) {
            $state = new State($billing_data->id_state);
            $xmlCustomer->appendChild($domtree->createElement('CustomerProvince', $state->iso_code));
        }

        $isoCountry = Country::getIsoById($billing_data->id_country);

        if ($isoCountry != null && !empty($isoCountry)) {
            $xmlCustomer->appendChild($domtree->createElement('CustomerCountry', $isoCountry));
        }

        if (!empty($billing_data->dni)) {
            $customer_fiscalcode = $domtree->createElement('CustomerFiscalCode');
            $customer_fiscalcode->appendChild($domtree->createCDataSection(Tools::strtoupper($billing_data->dni)));
            $xmlCustomer->appendChild($customer_fiscalcode);
        }

        if (!empty($billing_data->vat_number)) {
            $customerVatNumber = $this->cleanVatNumber($isoCountry, $billing_data->vat_number);
            $customer_vatcode = $domtree->CreateElement('CustomerVatCode');
            $customer_vatcode->appendChild($domtree->createCDataSection($customerVatNumber));
            $xmlCustomer->appendChild($customer_vatcode);
        }

        if (!empty($billing_data->phone) || !empty($billing_data->phone_mobile)) {
            $customer_cell_phone = $domtree->createElement('CustomerCellPhone');

            if (!empty($billing_data->phone)) {
                $customer_cell_phone->appendChild($domtree->createCDataSection($billing_data->phone));
            } elseif (!empty($billing_data->phone_mobile)) {
                $customer_cell_phone->appendChild($domtree->createCDataSection($billing_data->phone_mobile));
            }

            $xmlCustomer->appendChild($customer_cell_phone);
        }

        $customer_email = $domtree->createElement('CustomerEmail');
        $customer_email->appendChild($domtree->createCDataSection($customer->email));
        $xmlCustomer->appendChild($customer_email);

        return $domtree->saveXML();
    }

    // crea fattura xml
    public function createInvoiceXml($order_id, $isInvoice)
    {
        $domtree = new DOMDocument('1.0', 'UTF-8');
        $domtree->preserveWhiteSpace = false;
        $domtree->formatOutput = true;
        $xmlRoot = $domtree->createElement('Fattura24');
        $xmlRoot = $domtree->appendChild($xmlRoot);
        $fattura_invoice = $domtree->createElement('Document');
        $fattura_invoice = $xmlRoot->appendChild($fattura_invoice);

        $order = new Order($order_id);
        $this->trace('order details', $order);

        $invoice_address = $order->id_address_invoice;
        $billing_data = new Address($invoice_address);
        $customer = new Customer($billing_data->id_customer);

        /**
         * Aggiungo i campi per la fattura elettronica
         * Davide Iandoli 8.03.2019
         */
        $sql = 'select id_fattura24,fattura24_codice_destinatario,fattura24_pec from `' . _DB_PREFIX_;
        $sql .= 'fattura24` where id_customer = ' . (int) $order->id_customer . '';
        $result = Db::getInstance()->getRow($sql);
        $pec = $result ? $result['fattura24_pec'] : '';
        $isoCountry = Country::getIsoById($billing_data->id_country);

        if ($isoCountry != 'IT') {
            $sdi_code = 'XXXXXXX';
        } elseif (empty($result['fattura24_codice_destinatario'])) {
            $sdi_code = '0000000';
        } else {
            $sdi_code = $result['fattura24_codice_destinatario'];
        }

        $fattElChoice = Configuration::get('PS_CREAZIONE_FATTURA') == 1; // selezione fattura elettronica
        $disReceipt = Configuration::get('PS_DISABILITA_RICEVUTE') == 1; // selezione casella di controllo
        // creo una fattura elettronica in una di queste due condizioni
        $fattEl = $fattElChoice && !empty($billing_data->vat_number) || $disReceipt ? true : false;

        $customer_fePec = $domtree->createElement('FeCustomerPec');
        $customer_fePec->appendChild($domtree->createCDataSection($pec));
        $fattura_invoice->appendChild($customer_fePec);

        $customer_feCodiceDestinatario = $domtree->createElement('FeDestinationCode');
        $customer_feCodiceDestinatario->appendChild($domtree->createCDataSection($sdi_code));
        $fattura_invoice->appendChild($customer_feCodiceDestinatario);

        $company = empty($customer->company) ? trim($billing_data->company) : trim($customer->company);

        if (!empty($company)) {
            $customername = $company;
        } else {
            $customername = $billing_data->firstname . ' ' . $billing_data->lastname;
            /*$customername = empty($customer) ? $billing_data->firstname . ' ' . $billing_data->lastname :
                        $customer->firstname . ' ' . $customer->lastname;*/
        }

        $customer_name = $domtree->createElement('CustomerName');
        $customer_name->appendChild($domtree->createCDataSection($customername));
        $fattura_invoice->appendChild($customer_name);
        $customer_address = $domtree->createElement('CustomerAddress');
        $customer_address2 = $billing_data->address2 == null ? '' : ' ' . $billing_data->address2;
        $customer_address->appendChild($domtree->createCDataSection($billing_data->address1 . $customer_address2));
        $fattura_invoice->appendChild($customer_address);
        $fattura_invoice->appendChild($domtree->createElement('CustomerPostcode', $billing_data->postcode));
        $customer_name = $domtree->createElement('CustomerCity');
        $customer_name->appendChild($domtree->createCDataSection($billing_data->city));
        $fattura_invoice->appendChild($customer_name);
        if (!empty($billing_data->id_state)) {
            $state = new State($billing_data->id_state);
            $fattura_invoice->appendChild($domtree->createElement('CustomerProvince', $state->iso_code));
        } else {
            $fattura_invoice->appendChild($domtree->createElement('CustomerProvince', ''));
        }

        $isoCountry = Country::getIsoById($billing_data->id_country);
        $fattura_invoice->appendChild($domtree->createElement(
            'CustomerCountry',
            Country::getIsoById($billing_data->id_country)
        ));

        if (!empty($billing_data->dni)) {
            $customer_fiscalcode = $domtree->createElement('CustomerFiscalCode');
            $customer_fiscalcode->appendChild($domtree->createCDataSection(Tools::strtoupper($billing_data->dni)));
            $fattura_invoice->appendChild($customer_fiscalcode);
        }

        if (!empty($billing_data->vat_number)) {
            $customerVatNumber = $this->cleanVatNumber($isoCountry, $billing_data->vat_number);
            $customer_vatcode = $domtree->createElement('CustomerVatCode');
            $customer_vatcode->appendChild($domtree->createCDataSection($customerVatNumber));
            $fattura_invoice->appendChild($customer_vatcode);
        } elseif ($isoCountry !== 'IT' && empty($billing_data->vat_number) && $fattEl) {
            $customer_vatcode = $domtree->createElement('CustomerVatCode');
            $customer_vatcode->appendChild($domtree->createCDataSection($customername));
            $fattura_invoice->appendChild($customer_vatcode);
        }

        if (!empty($billing_data->phone) || !empty($billing_data->phone_mobile)) {
            $customer_cell_phone = $domtree->createElement('CustomerCellPhone');

            if (!empty($billing_data->phone)) {
                $customer_cell_phone->appendChild($domtree->createCDataSection($billing_data->phone));
            } elseif (!empty($billing_data->phone_mobile)) {
                $customer_cell_phone->appendChild($domtree->createCDataSection($billing_data->phone_mobile));
            }

            $fattura_invoice->appendChild($customer_cell_phone);
        }

        $customer_email = $domtree->createElement('CustomerEmail');
        $customer_email->appendChild($domtree->createCDataSection($customer->email));
        $fattura_invoice->appendChild($customer_email);

        if (!$order->isVirtual() && $order->id_address_delivery != null) {
            $delivery_address = new Address($order->id_address_delivery);
            if (!empty($delivery_address->company)) {
                $customerDeliveryName = $delivery_address->company;
            } else {
                $customerDeliveryName = $delivery_address->firstname . ' ' . $delivery_address->lastname;
            }
            $delivery_name = $domtree->createElement('DeliveryName');
            $delivery_name->appendChild($domtree->createCDataSection($customerDeliveryName));
            $fattura_invoice->appendChild($delivery_name);
            $address = $domtree->createElement('DeliveryAddress');
            $delivery_address2 = $delivery_address->address2 == null ? '' : ' ' . $delivery_address->address2;
            $address->appendChild($domtree->createCDataSection($delivery_address->address1 . $delivery_address2));
            $fattura_invoice->appendChild($address);
            $fattura_invoice->appendChild($domtree->createElement('DeliveryPostcode', $delivery_address->postcode));
            $fattura_invoice->appendChild($domtree->createElement('DeliveryCity', $delivery_address->city));
            if (!empty($delivery_address->id_state)) {
                $state = new State($delivery_address->id_state); // deve cercare l'indirizzo nel delivery_address
                $fattura_invoice->appendChild($domtree->createElement('DeliveryProvince', $state->iso_code));
            }
            $delivery_country = Country::getIsoById($delivery_address->id_country);
            $fattura_invoice->appendChild($domtree->createElement('DeliveryCountry', $delivery_country));
        }

        $TotalWithoutTax = number_format($order->total_paid_tax_excl, 2, '.', '');
        $fattura_invoice->appendChild($domtree->createElement('TotalWithoutTax', $TotalWithoutTax));
        $PS_payment_method = (object) Module::getInstanceByName($order->module);
        $payment_method_name = $domtree->createElement('PaymentMethodName');
        $orderPayment = isset($PS_payment_method->displayName) ? $PS_payment_method->displayName : '';
        $orderTotDiscountTaxExcl = (float) $order->total_discounts_tax_excl;
        $totShippingTaxExcl = (float) $order->total_shipping_tax_excl;
        $shippingRate = (float) number_format($order->carrier_tax_rate, 0, '.', '');

        /*
         * Con $payment_method_needle gestisco i pagamenti elettronici e lo stato Pagato della fattura
         * in conformità con la configurazione del modulo F24 - Davide Iandoli 20.01.2020
         */
        if (!empty($orderPayment) && strpos(Tools::strtolower($orderPayment), 'paypal') !== false) {
            $payment_method_needle = 'paypal';
        } elseif (strpos(Tools::strtolower($orderPayment), 'amzpayments') !== false) {
            $payment_method_needle = 'amazon payments';
        } elseif (strpos(Tools::strtolower($orderPayment), 'braintree') !== false) {
            $payment_method_needle = 'braintree';
        } elseif (strpos(Tools::strtolower($orderPayment), 'payplug') !== false) {
            $payment_method_needle = 'payplug';
        } elseif (strpos(Tools::strtolower($orderPayment), 'maofreepostepay') !== false) {
            $payment_method_needle = 'postepay';
        } elseif (strpos(Tools::strtolower($orderPayment), 'stripe') !== false) {
            $payment_method_needle = 'stripe';
        } elseif (strpos(Tools::strtolower($orderPayment), 'payment') !== false) {
            $payment_method_needle = 'pagamento smart o con carta';
        } elseif (strpos(Tools::strtolower($orderPayment), 'wallet') !== false) {
            $payment_method_needle = 'pagamento smart o con carta';
        } elseif (strpos(Tools::strtolower($orderPayment), 'pay') !== false) {
            $payment_method_needle = 'pagamento smart o con carta';
        } elseif (strpos(Tools::strtolower($orderPayment), 'prestashop checkout') !== false) {
            $payment_method_needle = 'pagamento smart o con carta';
        } else {
            $payment_method_needle = '';
        }


        $modulePayment = isset($PS_payment_method->name) ? $PS_payment_method->name : '';

        if (strpos($modulePayment, 'wire') !== false) { // ps 1.7 => 'ps_wirepayment'; ps 1.6 => 'bankwire';
            $FePaymentCode = 'MP05';
            $bank_name = $PS_payment_method->owner;
            $payment_method_name = $domtree->createElement('PaymentMethodName');
            $payment_method_name->appendChild($domtree->createCDataSection($bank_name));
            $fattura_invoice->appendChild($payment_method_name); // aggiungo il tag xml alla fattura
            $bank_detail = $PS_payment_method->details; // iban
            $payment_method_description = $domtree->createElement('PaymentMethodDescription');
            $payment_method_description->appendChild($domtree->createCDataSection($bank_detail));
            $fattura_invoice->appendChild($payment_method_description); // aggiungo il tag xml alla fattura
        } elseif ($modulePayment == 'ps_checkpayment') {    // Aggiunto pagamento con assegno
            $FePaymentCode = 'MP02';
            $payment_method_name->appendChild($domtree->createCDataSection($orderPayment));
            $fattura_invoice->appendChild($payment_method_name);
            $payment_method_description = $domtree->createElement('PaymentMethodDescription');
            $payment_method_description->appendChild($domtree->createCDataSection('Pagamento con assegno'));
            $fattura_invoice->appendChild($payment_method_description);
        } elseif ($modulePayment == 'ps_cashondelivery') {
            $FePaymentCode = 'MP01';
            $payment_method_name->appendChild($domtree->createCDataSection($orderPayment));
            $fattura_invoice->appendChild($payment_method_name);
            $payment_method_description = $domtree->createElement('PaymentMethodDescription');
            $payment_method_description->appendChild($domtree->createCDataSection('Pagamento alla consegna'));
            $fattura_invoice->appendChild($payment_method_description);
        } else {
            $uc_modulePayment = Tools::ucfirst($modulePayment);
            $payment_detail = empty($payment_method_needle) ?
                        'Pagamento smart o con carta' : Tools::ucfirst($payment_method_needle);
            $FePaymentCode = 'MP08';
            $payment_method_name->appendChild($domtree->createCDataSection($uc_modulePayment));
            $fattura_invoice->appendChild($payment_method_name);
            $payment_method_description = $domtree->createElement('PaymentMethodDescription');
            $payment_method_description->appendChild($domtree->createCDataSection($payment_detail));
            $fattura_invoice->appendChild($payment_method_description);
        }

        $fattura_invoice->appendChild($payment_method_description); // aggiungo il tag sempre
        if ($isInvoice) {
            $fattura_invoice->appendChild($domtree->createElement('FePaymentCode', $FePaymentCode));
        }

        $Total = number_format($order->total_paid, 2, '.', '');
        $vatAmount = $order->total_paid_tax_incl - $order->total_paid_tax_excl;
        $fattura_invoice->appendChild($domtree->createElement('VatAmount', $vatAmount));
        $fattura_invoice->appendChild($domtree->createElement('Total', $Total));

        if ($isInvoice) {
            $payments = $fattura_invoice->appendChild($domtree->createElement('Payments'));
            $payment = $payments->appendChild($domtree->createElement('Payment'));
            $payment->appendChild($domtree->createElement('Date', $this->now('Y-m-d')));
            $Amount = number_format($order->total_paid, 2, '.', ''); // arrotondamento a 2 decimali
            $payment->appendChild($domtree->createElement('Amount', $Amount));
            if (Configuration::get('PS_STATO_PAGATO') == 1 || (float) $order->total_paid == 0.00) {
                $payment->appendChild($domtree->createElement('Paid', 'true'));
            } elseif (Configuration::get('PS_STATO_PAGATO') == 2) {
                $modulePayment = $PS_payment_method->name;
                $bool_invoice_paid_status = !empty($payment_method_needle) ? 'true' : 'false';
                $payment->appendchild($domtree->createElement('Paid', $bool_invoice_paid_status));
            } else {
                $payment->appendChild($domtree->createElement('Paid', 'false'));
            }
        }

        /**
         * Il carrello mi servirà per controllare le restrizioni
         * nell'applicazione dello sconto
         */
        $cart = new Cart($order->id_cart);
        $f24CartRules = []; //array di regole carrello
        $giftProductArray = []; // array di prodotti omaggio
        $giftProductList = [];
        $cartRuleName = '';
        $couponArray = $order->getCartRules();
        $nCoupons = count($couponArray);

        /**
         * Qui creo un array personalizzato contenente tutti i codici sconto utilizzati
         * le chiavi reduction_percent e reduction_amount a loro volta contengono un array di info
         * che mi serviranno nelle righe prodotto
         */
        foreach ($couponArray as $cart_rule) {
            //$this->trace(' cart rule :', $cart_rule);
            $cartRuleName = $cart_rule['name'];
            $cartRuleDetails = new CartRule($cart_rule['id_cart_rule']);
            //$this->trace('cartRuleDetails :', $cartRuleDetails);
            $free_shipping = (int) $cartRuleDetails->free_shipping == 1;
            $giftProductId = (int) $cartRuleDetails->gift_product;
            $productRestrictionsArray = [];
            $productRestrictionsId = '';
            $reductionPercent = 0;
            $reductionAmount = (float) $cartRuleDetails->reduction_amount;
            $isAmountTaxExcl = (int) $cartRuleDetails->reduction_tax == 0;
            $valueTaxExcl = (float) $cart_rule['value_tax_excl'];
            $reductionProduct = (int) $cartRuleDetails->reduction_product;

            if ($giftProductId !== 0) {
                $giftProductList[] = $giftProductId;
            }

            $type = 'none';
            if ((float) $cartRuleDetails->reduction_percent > 0 &&
                                          ($reductionProduct == 0 || $reductionProduct == -2)) {
                $type = 'percent';
                $reductionPercent = (float) $cartRuleDetails->reduction_percent;
            } elseif ((float) $cartRuleDetails->reduction_percent > 0 && $reductionProduct == -1) {
                $type = 'percent_to_cheapest';
                $reductionPercent = (float) $cartRuleDetails->reduction_percent;
            } elseif ((float) $cartRuleDetails->reduction_percent > 0 && $reductionProduct > 0) {
                $type = 'percent_to_product';
                $reductionPercent = (float) $cartRuleDetails->reduction_percent;
            } elseif ((float) $cartRuleDetails->reduction_amount > 0) {
                $type = 'fixed';
            }

            // se lo sconto è fisso devo sottrarre il prezzo del prod omaggio


            // se necessario sottraggo l'importo della spedizione
            if ($free_shipping) {
                $valueTaxExcl -= $totShippingTaxExcl;
            }

            if (is_callable(array($cartRuleDetails, 'checkProductRestrictionsFromCart'))) {
                $productRestrictionsArray = $cartRuleDetails->checkProductRestrictionsFromCart($cart, true, true, true);

                /**
                 * copia dell'array contenente l'id del prodotto, usato per contollare l'elenco dei prodotti
                 * a cui va aggiunto lo sconto
                 */
                $productRestrictionsId = array_map([$this, 'getParamSubstring'], $productRestrictionsArray);
            }

            $f24CartRules[] = [
                    'name' => $cartRuleName,
                    'type' => $type,
                    'free_shipping' => $free_shipping,
                    'gift_product' => $giftProductId,
                    'product_ids' => $productRestrictionsId,
                    'reduction_product' => $reductionProduct,
                    'reduction_amount' => $reductionAmount,
                    'is_amount_tax_excl' => $isAmountTaxExcl,
                    'reduction_percent' => $reductionPercent,
                    'value_tax_excl' => $valueTaxExcl
                ];
        }
        $this->trace('f24 cart rules :', $f24CartRules);

        $free_shipping = in_array(true, array_column($f24CartRules, 'free_shipping'));

        // sottraggo il totale spedizione dal totale sconto
        if ($free_shipping) {
            $orderTotDiscountTaxExcl -= $totShippingTaxExcl;
        }

        $products = $fattura_invoice->appendChild($domtree->createElement('Rows'));

        $productArray = $order->getProducts();
        $productList = [];
        $itemTax = [];
        $cartDiscount = []; // array di sconti fissi a carrello
        $orderTaxes = $order->getOrderDetailTaxes();

        foreach ($productArray as $product_detail) {
            $this->trace('item details :', $product_detail);

            $idPdc = Configuration::get('PS_F24_PDC');
            if (!empty($idPdc) && $idPdc != 'Nessun Pdc') {
                $pdcSet = true;
            }

            $productId = (int) $product_detail['product_id'];
            $qty = (int) $product_detail['product_quantity'];
            $productDescription = $product_detail['product_name'];
            $price = (float) $product_detail['unit_price_tax_excl'];
            $discountAmount = 0;
            $isGift = false;
            $giftPrice = 0;
            $productCode = $product_detail['product_reference'];

            foreach ($giftProductList as $item) {
                if ($productId == $item) {
                    $qty -=1;
                    $orderTotDiscountTaxExcl -= $price;
                    $giftProductArray[] = $product_detail;
                    $giftPrice += $price;
                    $isGift = true;
                }
            }

            $idTaxRule = (int) $product_detail['id_tax_rules_group'];
            $idOrderDetail = (int) $product_detail['id_order_detail'];
            $taxId = $this->getTaxId($orderTaxes, $idTaxRule, $idOrderDetail);
            $taxRateFromDetail = (float) $product_detail['tax_rate'];
            $taxRateFromId = $this->getTaxRateFromId($taxId);
            $taxRate = $taxRateFromDetail == $taxRateFromId ? $taxRateFromDetail : $taxRateFromId;
            $taxNameFromId = $this->getTaxNameFromId($taxId);
            $natura = '';

            /**
             * Controllo solo il rate per agganciare l'aliquota IVA e non crearne una nuova
             * anche nel caso in cui il cliente voglia creare solo un Ordine
             * In questo modo potrà trasformarlo in FE senza un errore di mancata indicazione natura
             */
            if ((float)$taxRate == 0) {
                $natura = $this->getNaturaByIdTax($idTaxRule);
            }


            $taxName = $product_detail['tax_calculator']->getTaxesName();
            //$VatDescription = !empty($taxName) ? $taxName : $this->getDefaultVatDescription($taxRate);
            $VatDescription = !empty($taxName) ? $taxName : $taxNameFromId;

            foreach ($f24CartRules as $rule) {
                $isAmountTaxExcl = $rule['is_amount_tax_excl'];

                if ($free_shipping) {
                    $rule['value_tax_excl'] -= $totShippingTaxExcl;
                }

                // prodotto incluso in sconto percentuale
                $productIncluded = in_array($productId, $rule['product_ids']);
                /**
                 * Se ci sono più sconti in percentuale devono essere
                 * applicati 'a cascata', quindi devo sottrarre al prezzo
                 * del prodotto l'importo dello sconto precedente
                 * si noti che se c'è uno sconto solo discountAmount è
                 * uguale a zero
                 *
                 * Se lo sconto è percentuale non devo tener conto del prezzo del prodotto
                 * omaggio, perché ho già cambiato la quantità
                 */
                if ($rule['type'] == 'percent' && $productIncluded) {
                    $newPrice = ($price * $qty) - $discountAmount;
                    $discountAmount += (float) number_format(($newPrice / 100
                                             * (float) $rule['reduction_percent']), 8, '.', '');
                } elseif ($rule['type'] == 'percent_to_product' && $productId == $reductionProduct) {
                    $discountAmount += (float) number_format((($price * $qty) / 100
                                             * (float) $rule['reduction_percent']), 8, '.', '');
                }
                //$this->trace('discount amount 1 :', $discountAmount);
            }

            /**
             * Qui aggiungo gli sconti fissi ristretti al prodotto
             * cui devo sottrarre il prezzo del prodotto omaggio
             */
            $productReductionAmount = $this->f24getRedAmount($productId, $f24CartRules);
            $discountAmount += $productReductionAmount;
            //$this->trace('discount amount 2 :', $discountAmount);


            $productList[] = array(
                               'id' => $productId,
                               'code' => $productCode,
                               'description' => $productDescription,
                               'price' => $price,
                               'qty' => $qty,
                               'is_gift' => $isGift,
                               'tax_rate' => $taxRate,
                               'tax_description' => $VatDescription,
                               'natura' => $natura,
                               'disc_amount' => $discountAmount

            );

            if (!array_key_exists((int) $taxRate, $itemTax) && $discountAmount > 0) {
                $itemTax[$taxRate] = ['vat_description' => $VatDescription,
                                      'natura' => $natura,
                                      'total_price_per_vat_rate' => $price,
                                      'total_discount_per_vat_rate' => $discountAmount];
            } elseif ($discountAmount > 0) {
                $itemTax[$taxRate]['total_price_per_vat_rate'] += $price;
                $itemTax[$taxRate]['total_discount_per_vat_rate'] += $discountAmount;
            }
            $this->trace('item tax :', $itemTax);
            // sconti che si applicano a carrello
            if ($orderTotDiscountTaxExcl > 0) {
                if (!array_key_exists((int) $taxRate, $cartDiscount)) {
                    $cartDiscount[$taxRate] = [
                                           'vat_description' => $VatDescription,
                                           'natura' => $natura,
                                           'total_price_per_vat_rate' => $price,
                                           'total_discount_per_vat_rate' => 0
                                        ];
                } else {
                    $cartDiscount[$taxRate]['total_price_per_vat_rate'] += $price;
                }
            }
            $this->trace('Cart Discount :', $cartDiscount);
        }

        $this->addProductRowsToXml($domtree, $products, $productList);


        if (!empty($giftProductArray)) {
            $this->addGiftProductToXml($domtree, $products, $giftProductArray, $f24CartRules, $orderTaxes);
        }

        if ($nCoupons > 0) {
            $itemTax = $this->updateItemTax(
                $itemTax,
                $cartDiscount,
                $orderTotDiscountTaxExcl
            );
        }

        if (!empty($itemTax)) {
            $this->addDiscProduct($domtree, $products, $itemTax);
        }

        /**
        * Gestione riga di spedizione, al termine delle righe prodotto
        */
        if ($order->total_shipping_tax_excl >= 0) {
            $cartRuleName = implode("\n", array_column($f24CartRules, 'name'));
            $pdcSet = false;
            $idPdc = Configuration::get('PS_F24_PDC');
            if (!empty($idPdc) && $idPdc != 'Nessun Pdc') {
                $pdcSet = true;
            }

            //$products = $fattura_invoice->appendChild($domtree->createElement('Rows'));

            $sqlRow = "SELECT `name` FROM `" . _DB_PREFIX_ . "carrier` 
                       WHERE `id_carrier` = '" . pSQL($order->id_carrier) . "'";
            $shipping_company = Db::getInstance()->getRow($sqlRow);

            $billing_data = new Address($order->id_address_invoice);
            $carrier = new Carrier($order->id_carrier);
            $VatDescription = $carrier->getTaxCalculator($billing_data)->getTaxesName();

            $shippingLabel = $this->l('Shipping amount :');
            $shippingDescription = !$free_shipping ? $shippingLabel . ' ' . $shipping_company['name'] :
                                    $shippingLabel . ' ' . $shipping_company['name'] . ' - ' . $cartRuleName;

            $product = $products->appendChild($domtree->createElement('Row'));
            $description = $domtree->createElement('Description');

            $description->appendChild($domtree->createCDataSection($shippingDescription));
            $product->appendChild($description);
            $product->appendChild($domtree->createElement('Qty', '1'));
            $product->appendChild($domtree->createElement('Price', $order->total_shipping_tax_excl));
            if ($free_shipping) {
                $product->appendChild($domtree->createElement('Discounts', 100));
            }
            //$shippingRate = number_format($order->carrier_tax_rate, 0, '.', '');
            $product->appendChild($domtree->createElement('VatCode', $shippingRate));


            $naturaSpedizione = Configuration::get('PS_CONF_NATURA_SPEDIZIONE');
            $lista = $this->getListaNatureNew();
            $needle = Tools::substr($naturaSpedizione, 0, 2);

            /*
            * Con questo blocco gestisco i casi in cui l'aliquota IVA
            * per la spedizione è 0%, inserendo la Natura
            */
            if ((int) $shippingRate == 0) {
                $naturaSpedizione = Configuration::get('PS_CONF_NATURA_SPEDIZIONE');

                $lista = $this->getListaNatureNew(); // uso il nuovo metodo
                $needle = Tools::substr($naturaSpedizione, 0, 2);

                if ($naturaSpedizione == 'N0') {
                    $naturaSpedizione = '';
                    $VatDescription = 'IVA';
                } elseif (is_array($lista[$needle])) {
                    $VatDescription = $lista[$needle][$naturaSpedizione];
                } else {
                    $VatDescription = $lista[$naturaSpedizione];
                }
            }

            $product->appendChild($domtree->createElement('VatDescription', $VatDescription));

            if ($shippingRate == 0) {
                $product->appendChild($domtree->createElement('FeVatNature', $naturaSpedizione));
            }
            if ($pdcSet) {
                $product->appendChild($domtree->createElement('IdPdc', $idPdc));
            }
        }


        if ($isInvoice) {
            $footnotes = $domtree->createElement('FootNotes');
            $footnotes->appendChild($domtree->CreateCDataSection($this->l('Order ref. : ') . $order->reference));
            $fattura_invoice->appendChild($footnotes);
            $f24_inv_object = Configuration::get('PS_INV_OBJECT');
            $f24_search = ['(N)', '(R)'];
            $f24_replace = [$order->id, $order->reference];
            $f24_send_object = str_replace($f24_search, $f24_replace, $f24_inv_object);
            $f24_xml_object = $domtree->createElement('Object');
            $f24_xml_object->appendChild($domtree->createCDataSection($f24_send_object));
            $fattura_invoice->appendChild($f24_xml_object);

            if (empty($billing_data->vat_number)) {
                if (Configuration::get('PS_CREAZIONE_FATTURA') == 3) {
                    $documentType = 'R';
                }
                if (Configuration::get('PS_CREAZIONE_FATTURA') == 2) {
                    $documentType = Configuration::get('PS_DISABILITA_RICEVUTE') == 1 ? 'I-force' : 'R';
                }
                if (Configuration::get('PS_CREAZIONE_FATTURA') == 1) {
                    $documentType = Configuration::get('PS_DISABILITA_RICEVUTE') == 1 ? 'FE' : 'R'; // create fattura
                }
            } else {
                if (Configuration::get('PS_CREAZIONE_FATTURA') == 3) {
                    $documentType = 'R';
                }
                if (Configuration::get('PS_CREAZIONE_FATTURA') == 2) {
                    $documentType = 'I';
                }
                if (Configuration::get('PS_CREAZIONE_FATTURA') == 1) {
                    $documentType = 'FE';
                }
            }
            //se il cliente non ha una partita iva uso il sezionale ricevuta
            if ($documentType == 'R') { // use receipts issue number bugfix 30.01.2020
                $idNumerator = Configuration::get('PS_F24_SEZIONALE_RICEVUTA');
            } else { // uso il sezionale fattura
                $idNumerator = Configuration::get('PS_F24_SEZIONALE_FATTURA');
            }

            if ($idNumerator != 'Predefinito') {
                $fattura_invoice->appendChild($domtree->createElement('IdNumerator', $idNumerator));
            }

            $query = 'SELECT docIdOrderFattura24 FROM ' . _DB_PREFIX_ . 'orders WHERE id_order=\'' .
                    (int) $order_id . '\';';
            $docIdOrderFattura24 = Db::getInstance(_PS_USE_SQL_SLAVE_)->getRow($query)['docIdOrderFattura24'];
            if ($docIdOrderFattura24 != null) {
                $fattura_invoice->appendChild($domtree->createElement('F24OrderId', $docIdOrderFattura24));
            }
            $sendEmail = Configuration::get('PS_EMAIL_FATTURA') == 1 ? 'true' : 'false';
            //$updateStorage = Configuration::get('PS_F24_MAGAZZINO_FATTURA');

            if ($order->isVirtual()) {
                $idTemplate = Configuration::get('PS_F24_MODELLO_FATTURA');
            } else {
                $idTemplate = Configuration::get('PS_F24_MODELLO_FATTURA_DEST');
            }
        } else {
            $orderNumberOption = Configuration::get('PS_NUMERO_ORDINE');
            if ($orderNumberOption == 0) {
                $f24OrderNumber = $order->id;
            } elseif ($orderNumberOption == 1) {
                $f24OrderNumber = $order->reference;
            } elseif ($orderNumberOption == 2) {
                $f24OrderNumber = $order->id . ' - ' . $order->reference;
            }

            $fattura_invoice->appendChild($domtree->createElement('Number', $f24OrderNumber));

            $documentType = 'C';
            $sendEmail = Configuration::get('PS_MAIL_ORDINE') == 1 ? 'true' : 'false';
            //$updateStorage = Configuration::get('PS_F24_MAGAZZINO_ORDINE');
            if ($order->isVirtual()) {
                $idTemplate = Configuration::get('PS_F24_MODELLO_ORDINE');
            } else {
                $idTemplate = Configuration::get('PS_F24_MODELLO_ORDINE_DEST');
            }
        }
        $fattura_invoice->appendChild($domtree->createElement('DocumentType', $documentType));

        /**
         * Con questo blocco riporto il bollo virtuale per le FE
         * e scrivo il tag solo se il valore è "V"
         * tre condizioni: opzione selezionata + tipo documento + totale > 77.47
         * Davide Iandoli 16.04.2020
         */
        $bollo_option = Configuration::get('PS_F24_BOLLO_VIRTUALE_FE');
        if ($bollo_option == 1 && $documentType == 'FE' && $Total > 77.47) {
            $bollo_fe = 'V';
            $fattura_invoice->appendChild($domtree->createElement('FeVirtualStamp', $bollo_fe));
        } else {
            $bollo_fe = 'N';
        }

        $fattura_invoice->appendChild($domtree->createElement('SendEmail', $sendEmail));

        if (!empty($idTemplate) && $idTemplate !== 'Predefinito') {
            $fattura_invoice->appendChild($domtree->createElement('IdTemplate', $idTemplate));
        }
        return $domtree->saveXML();
    }

    // download documento da F24
    public function downloadDocument($docIdFattura24, $orderId, $isInvoice)
    {
        $fattura24_api_url = $this->baseUrl . 'GetFile';
        $command = $this->getApiCommand($fattura24_api_url); // ottengo il comando
        $start_time = $this->startTime();
        $send_data = [];
        $send_data['apiKey'] = Configuration::get('PS_FATTURA24_API');
        $send_data['source'] = $this->sourceInfo;
        $send_data['docId'] = (int) $docIdFattura24;
        $dataReturned = $this->curlDownload($fattura24_api_url, http_build_query($send_data));
        //$this->trace('data returned :', $dataReturned);
        $finish_time = $this->endTime();
        $this->getApiTime($command, $start_time, $finish_time);
        libxml_use_internal_errors(true);
        if (!simplexml_load_string(utf8_encode($dataReturned))) {
            if ($isInvoice) {
                $order = new Order((int) $orderId);
                //$this->trace('invoice date :', $order->date_upd);
                // la proprietà date_upd corrisponde all'ultimo aggiornamento dell'ordine
                $fileName = $this->constructFileName($order->invoice_number, $order->date_upd);
            } else {
                $fileName = 'ORD' . sprintf('%06d', (int) $orderId) . '.pdf';
            }
            file_put_contents(_PS_DOWNLOAD_DIR_ . 'Fattura24/' . $fileName, $dataReturned);
        }
    }

    // salva contatto in F24
    public function saveCustomer($check_order)
    {
        $fattura24_api_url = $this->baseUrl . 'SaveCustomer';
        $command = $this->getApiCommand($fattura24_api_url); // ottengo il comando
        $start_time = $this->startTime();
        $send_data = [];
        $send_data['apiKey'] = Configuration::get('PS_FATTURA24_API');
        $send_data['source'] = $this->sourceInfo;
        $send_data['xml'] = $this->createCustomerXml($check_order->id);
        $dataReturned = $this->curlDownload($fattura24_api_url, http_build_query($send_data));
        $finish_time = $this->endTime();
        $this->getApiTime($command, $start_time, $finish_time);

        $response = new DOMDocument();
        $response->preserveWhiteSpace = false;
        $response->formatOutput = true;
        $response->loadXML($dataReturned);
        $response_string = $response->saveXML();
        $this->trace($send_data['xml'], $response_string);
    }

    /**
     * Con queste funzioni calcolo il tempo di risposta
     * delle chiamate API, compatibilmente con la versione PHP in uso
     */
    public function startTime()
    {
        if (Tools::substr($this->phpversion, 0, 3) < '7.3') {
            $start_time = microtime(true);
        } else {
            $start_time = hrtime(true);
        }

        return $start_time;
    }

    public function endTime()
    {
        if (Tools::substr($this->phpversion, 0, 3) < '7.3') {
            $finish_time = microtime(true);
        } else {
            $finish_time = hrtime(true);
        }

        return $finish_time;
    }

    public function getApiCommand($command)
    {
        $api_command_array = explode('/', $command);

        return $api_command_array[5];
    }

    public function getApiTime($command, $start_time, $finish_time)
    {
        // formule per convertire in millisecondi
        if (Tools::substr($this->phpversion, 0, 3) < '7.3') {
            $duration = ($finish_time - $start_time) * 1000;
        } else {
            $duration = round(($finish_time - $start_time) / 10000000);
        }
        $result = number_format($duration, 0, '.', '');
        $response_time = ['chiamata :' => $command, 'tempo di risposta in millisecondi :' => $result];
        $this->trace('controllo risposte :', $response_time);
    }

    /*
     * Funzione comune da eseguire dopo gli hook dichiarati sopra,
     * il parametro isInvoice serve per distinguere se deve essere creato un ordine o una fattura
     */
    public function afterHook($order, $isInvoice)
    {
        $orderId = $order->id;
        $fattura24_api_url = $this->baseUrl . 'SaveDocument';
        $command = $this->getApiCommand($fattura24_api_url); // ottengo il comando
        $start_time = $this->startTime();

        $send_data = [];
        $send_data['apiKey'] = Configuration::get('PS_FATTURA24_API');
        $send_data['source'] = $this->sourceInfo;
        $send_data['xml'] = $this->createInvoiceXml($orderId, $isInvoice);
        
        if (!$isInvoice) {
            $idRequest = 'C' . $orderId;
        } else {
            $xml = simplexml_load_string($send_data['xml']);
            $docType = '';
            if (is_object($xml)) {
                $docType = $xml->Document->DocumentType;
            }
            $idRequest = $docType . $orderId;
        }
       
        $send_data['idRequest'] = $idRequest;
        $dataReturned = $this->curlDownload($fattura24_api_url, http_build_query($send_data));
        $finish_time = $this->endTime();
        $this->getApiTime($command, $start_time, $finish_time);
        $obj_dataReturned = simplexml_load_string(utf8_encode($dataReturned), 'SimpleXMLElement', LIBXML_NOCDATA);
        $json_dataReturned = json_encode($obj_dataReturned);

        if (!$isInvoice) {
            $column = 'apiResponseOrder';
            if (is_object($obj_dataReturned)) {
                $docIdOrderFattura24 = $obj_dataReturned->docId;
            //$this->trace('docId 1 :', $docIdOrderFattura24);
            } else {
                $docIdOrderFattura24 = explode('</docId>', explode('<docId>', $dataReturned)[1])[0];
                //$this->trace('docId 2 :', $docIdOrderFattura24);
            }

            $query = 'UPDATE ' . _DB_PREFIX_ . 'orders SET docIdOrderFattura24=\'' . pSQL((int) $docIdOrderFattura24);
            $query .= '\' WHERE id_order=\'' . (int) $orderId . '\';';
            Db::getInstance()->Execute($query);
            $this->downloadDocument($docIdOrderFattura24, $orderId, false);
        } else {
            $column = 'apiResponseInvoice';
            if (is_object($obj_dataReturned)) {
                $docIdFattura24 = $obj_dataReturned->docId;
            } else {
                //$this->trace('data returned :', $dataReturned);
                $docIdFattura24 = explode('</docId>', explode('<docId>', $dataReturned)[1])[0];
            }

            $query = 'UPDATE ' . _DB_PREFIX_ . 'orders SET docIdFattura24=\'' . pSQL((int) $docIdFattura24);
            $query .= '\' WHERE id_order=\'' . (int) $orderId . '\';';
            Db::getInstance()->Execute($query);
            //$this->trace('doc id fattura24 :', $docIdFattura24);

            $this->downloadDocument($docIdFattura24, $orderId, true);
        }

        /**
         *  Con questo blocco gestisco i casi in cui le colonne apiResponseOrder
         *  e apiResponseInvoice NON sono state create
         *  Davide Iandoli 19.08.2020
         */
        $sql = 'SELECT apiResponseOrder, apiResponseInvoice from ' . _DB_PREFIX_ . 'orders';
        $result = Db::getInstance()->ExecuteS($sql);

        if ($result) {
            if (is_object($obj_dataReturned)) {
                $query = 'UPDATE ' . _DB_PREFIX_ . 'orders SET ' . $column . '=\'' . pSQL((string) $json_dataReturned);
                $query .= '\' WHERE id_order=\'' . (int) $orderId . '\';';
                Db::getInstance()->Execute($query);
            } else {
                $query = 'UPDATE ' . _DB_PREFIX_ . 'orders SET ' . $column . '=\'' .
                        pSQL((string) $dataReturned);
                $query .= '\' WHERE id_order=\'' . (int) $orderId . '\';';
                Db::getInstance()->Execute($query);
            }
        }
        // fine blocco fix

        $response = new DOMDocument();
        $response->preserveWhiteSpace = false;
        $response->formatOutput = true;
        $response->loadXML($dataReturned);
        $response_string = $response->saveXML();
        $this->trace($send_data['xml'], $response_string);
    }

    public function f24Options()
    {
        $f24_fields = [
                        'PS_FATTURA24_API',
                        'PS_SALVA_CLIENTE',
                        'PS_CREA_ORDINE',
                        'PS_MAIL_ORDINE',
                        'PS_F24_MODELLO_ORDINE',
                        'PS_F24_MODELLO_ORDINE_DEST',
                        'PS_CREAZIONE_FATTURA',
                        'PS_CONF_NATURA_IVA',
                        'PS_CONF_NATURA_SPEDIZIONE',
                        'PS_INV_OBJECT',
                        'PS_EMAIL_FATTURA',
                        'PS_STATO_PAGATO',
                        'PS_DISABILITA_RICEVUTE',
                        'PS_F24_MODELLO_FATTURA',
                        'PS_F24_MODELLO_FATTURA_DEST',
                        'PS_F24_PDC',
                        'PS_F24_SEZIONALE_RICEVUTA',
                        'PS_F24_SEZIONALE_FATTURA',
                        'PS_F24_BOLLO_VIRTUALE_FE',
                ];
        return $f24_fields;
    }

    public function f24Value($param)
    {
        return [$param => Tools::getValue($param), Configuration::get($param)];
    }

    // opzioni 'Crea documento fiscale'
    public function getInvoiceOptions()
    {
        return [$this->l('Disabled'),
                     $this->l('Electronic invoice'),
                     $this->l('NON electronic Invoice'),
                     $this->l('NON electronic Receipt'), ];
    }

    // opzioni stato pagato
    public function f24PaidStatus()
    {
        return [$this->l('Never'),
                     $this->l('Always'),
                     $this->l('E-payments (e.g.: Paypal)'), ];
    }

    // opzioni 'bollo virtuale'
    public function f24StampOptions()
    {
        return [$this->l('Never'),
                $this->l('VAT exempt fiscal regime and order total > 77,47 Euros'), ];
    }

    // opzioni numero ordine
    public function f24OrderNumber()
    {
        return [
                $this->l('Prestashop order id'),
                $this->l('Prestashop order reference'),
                $this->l('Prestashop order id - order reference')
               ];
    }

    // lista modelli
    public function getTemplate($templates, $isOrder)
    {
        $listaNomi = [];
        $listaNomi['Predefinito'] = $this->l('Default');
        $xml = simplexml_load_string(utf8_encode($templates));
        if (is_object($xml)) {
            $listaModelli = $isOrder ? $xml->modelloOrdine : $xml->modelloFattura;
            foreach ($listaModelli as $modello) {
                $listaNomi[(int) ($modello->id)] = (string) ($modello->descrizione);
            }
        } else {
            if (!empty($templates)) {
                $this->trace('error list templates', $templates);
            }
        }

        return $listaNomi;
    }

    // lista Pdc
    public function getPdc($f24_pdc)
    {
        $listaNomi = [];
        $listaNomi['Nessun Pdc'] = $this->l('No Pdc');
        $xml = simplexml_load_string(utf8_encode($f24_pdc));
        if (is_object($xml)) {
            foreach ($xml->pdc as $pdc) {
                if ((int) ($pdc->ultimoLivello) == 1) {
                    $list = str_replace('^', '.', (string) ($pdc->codice)) . ' - ' . (string) ($pdc->descrizione);
                    $listaNomi[(int) ($pdc->id)] = $list;
                }
            }
        } else {
            if (!empty($f24_pdc)) {
                $this->trace('error list pdc', $f24_pdc);
            }
        }

        return $listaNomi;
    }

    // lista sezionali
    public function getNumerator($f24_sezionali, $idTipoDocumento)
    {
        $listaNomi = [];
        $listaNomi['Predefinito'] = $this->l('Default');
        $xml = simplexml_load_string(utf8_encode($f24_sezionali));
        if (is_object($xml)) {
            foreach ($xml->sezionale as $sezionale) {
                foreach ($sezionale->doc as $doc) {
                    if ((int) ($doc->id) == $idTipoDocumento && (int) ($doc->stato) == 1) {
                        $listaNomi[(int) ($sezionale->id)] = (string) ($sezionale->anteprima);
                    }
                }
            }
        } else {
            if (!empty($f24_sezionali)) {
                $this->trace('error list sezionale', $f24_sezionali);
            }
        }

        return $listaNomi;
    }

    /**
     * Con questo metodo restituisco un solo array di aliquote IVA
     * contenente tutti gli sconti: se sono presenti sconti a carrello
     * il risultato sarà aggiungere ai dati già presenti la quota parte di sconto
     * ripartendola per aliquota
     */
    public function updateItemTax(
        $itemTax,
        $cartDiscount,
        $orderTotDiscountTaxExcl
    ) {
        $maxDiscountAmount = $orderTotDiscountTaxExcl;
        $totalDiscounted = array_sum(array_column($itemTax, 'total_discount_per_vat_rate'));
        // formatto il risultato con due decimali per una corretta comparazione
        $totDiscResult = (float) number_format($totalDiscounted, 2, '.', '');
        //$this->trace('max discount Amount', $maxDiscountAmount);
        //$this->trace('total Discounted :', $totDiscResult);
        $this->trace('item tax iniziale :', $itemTax);

        /**
         * In caso in cui ci sia un prodotto omaggio
         * faccio questo controllo ulteriore per capire se devo
         * ricalcolare total_discount_per_vat_rate
         * Da notare che questa logica si applica solo a sconti con restrizioni sui prodotti
         */
        if ($totDiscResult > $maxDiscountAmount /*&& $giftPrice > 0*/) {
            $this->trace('condizione verificata :');
            foreach ($itemTax as $key => $value) {
                if ($itemTax[$key]['total_discount_per_vat_rate'] > 0) {
                    $itemTax[$key]['total_discount_per_vat_rate'] = $maxDiscountAmount;
                }
            }
        }
        $this->trace('item tax mediano :', $itemTax);

        $discountDiff = $orderTotDiscountTaxExcl - $totalDiscounted;
        $this->trace('differenza :', $discountDiff);

        if ($discountDiff > 0 && $discountDiff > 0.01) {
            // il totale prodotti scontati in questo caso è diverso
            $totPriceProdDisc = (float) array_sum(array_column($cartDiscount, 'total_price_per_vat_rate'));

            foreach ($cartDiscount as $key => $value) {
                $total_price_per_rate = (float) $value['total_price_per_vat_rate'];
                $x = ($total_price_per_rate * 100) / $totPriceProdDisc;
                $price = (float) number_format(($discountDiff * $x / 100), 8, '.', '');

                if (!array_key_exists((int) $key, $itemTax)) {
                    $itemTax[$key] = $cartDiscount[$key]; // aggiungo l'array già esistente
                    $itemTax[$key]['total_discount_per_vat_rate'] += $price;
                } else {
                    $itemTax[$key]['total_discount_per_vat_rate'] += $price;
                }
            }
        }
        $this->trace('item tax finale :', $itemTax);

        return $itemTax;
    }

    /**
     * Uso questo metodo per controllare se uno sconto fisso
     * è ristretto al prodotto - aggiunto controllo su tipo sconto
     * cfr ticket n.: 70957 - Davide Iandoli 11.02.2022
     */
    public function f24getRedAmount($productId, $f24CartRules)
    {
        $amountToAdd = 0;

        foreach ($f24CartRules as $rule) {
            $isPercent = $rule['type'] == 'percent';
            if ($rule['reduction_product'] == $productId && !$isPercent) {
                $amountToAdd += (float) $rule['value_tax_excl'];
            }
        }

        return $amountToAdd;
    }


    /**
     * Solo in caso di sconto a importo fisso chiamo questo metodo con cui intendo
     * creare una riga prodotto per ogni aliquota IVA, secondo la vecchia logica
     */
    public function addDiscProduct($domtree, $products, $itemTax)
    {
        $discount_name = 'Sconto '; // altrimenti la forzo

        $pdcSet = false;
        $idPdc = Configuration::get('PS_F24_PDC');
        if (!empty($idPdc) && $idPdc != 'Nessun Pdc') {
            $pdcSet = true;
        }

        foreach ($itemTax as $key => $value) {
            $price = $value['total_discount_per_vat_rate'];

            // non aggiungo righe con importo zero!
            if ((float) $price == 0.00) {
                break;
            }

            $product = $products->appendChild($domtree->createElement('Row'));
            $code = $domtree->createElement('Code');
            $code->appendChild($domtree->createCDataSection(''));
            $product->appendChild($code);
            $description = $domtree->createElement('Description');
            $description->appendChild($domtree->createCDataSection($discount_name . ' IVA ' . $key . '%'));
            $product->appendChild($description);
            $product->appendChild($domtree->createElement('Qty', 1));
            $product->appendChild($domtree->createElement('Price', number_format(-$price, 8, '.', '')));
            $product->appendChild($domtree->createElement('VatCode', $key));
            $product->appendChild($domtree->createElement('VatDescription', $value['vat_description']));
            if ($key == 0) {
                $product->appendChild($domtree->createElement('FeVatNature', $value['natura']));
            }

            if ($pdcSet) {
                $idPdc = Configuration::get('PS_F24_PDC');
                $product->appendChild($domtree->createElement('IdPdc', $idPdc));
            }
        }
    }

    /**
     * Aggiungo le righe prodotto: gestisco sconti in percentuale e prodotti omaggio
     * in questo ultimo caso devo sottrarre una unità alla quantità acquista se il prodotto
     * è lo stesso che ho messo nel carrello (es.: 2 quadri al prezzo di uno) perché
     * abbiamo deciso di crare una riga prodotto a parte per gli omaggi:
     * così diventa più facile gestire sconti multipli che abbiano sia omaggio
     * sia percentuale sul carrello, e la fattura diventa più trasparente per l'utente
     */
    public function addProductRowsToXml($domtree, $products, $productList)
    {
        $pdcSet = false;
        $idPdc = Configuration::get('PS_F24_PDC');
        if (!empty($idPdc) && $idPdc != 'Nessun Pdc') {
            $pdcSet = true;
        }

        foreach ($productList as $product_detail) {
            $this->trace('product list details', $product_detail);

            $qty = $product_detail['qty'];
            $productCode = $product_detail['code'];
            $productDescription = $product_detail['description'];
            $price = $product_detail['price'];
            $taxRate = (float) $product_detail['tax_rate'];
            $VatDescription = $product_detail['tax_description'];
            $natura = $product_detail['natura'];

            /**
             * se il prodotto è in omaggio la quantità è zero
             * ed esco dal ciclo
             */
            if ((int) $qty < 1 || $price == 0) {
                continue;
            }

            $product = $products->appendChild($domtree->createElement('Row'));
            $code = $domtree->createElement('Code');
            $code->appendChild($domtree->createCDataSection($productCode));
            $product->appendChild($code);
            $description = $domtree->createElement('Description');
            $description->appendChild($domtree->createCDataSection($productDescription));

            $product->appendChild($description);
            $product->appendChild($domtree->createElement('Qty', $qty));
            $product->appendChild($domtree->createElement('Price', $price));
            $product->appendChild($domtree->createElement('VatCode', $taxRate));
            $product->appendChild($domtree->createElement('VatDescription', $VatDescription));

            if ((float)$taxRate == 0) {
                $product->appendChild($domtree->createElement('FeVatNature', $natura));
            }

            if ($pdcSet) {
                $product->appendChild($domtree->createElement('IdPdc', $idPdc));
            }
        }
    }


    /**
     *  Gestione separata dei prodotti omaggio
     */
    public function addGiftProductToXml($domtree, $products, $giftProductArray, $f24CartRules, $orderTaxes)
    {
        $pdcSet = false;
        $idPdc = Configuration::get('PS_F24_PDC');
        if (!empty($idPdc) && $idPdc != 'Nessun Pdc') {
            $pdcSet = true;
        }

        foreach ($giftProductArray as $product_detail) {
            $cartRuleName = implode("\n", array_column($f24CartRules, 'name'));
            $productDescription = $product_detail['product_name']. "\n" . $cartRuleName;

            $product = $products->appendChild($domtree->createElement('Row'));
            $code = $domtree->createElement('Code');
            $code->appendChild($domtree->createCDataSection($product_detail['product_reference']));
            $product->appendChild($code);
            $description = $domtree->createElement('Description');
            $description->appendChild($domtree->createCDataSection($productDescription));

            $product->appendChild($description);
            $product->appendChild($domtree->createElement('Qty', 1));
            $product->appendChild($domtree->createElement('Price', $product_detail['unit_price_tax_excl']));
            $product->appendChild($domtree->createElement('Discounts', 100));

            $idTaxRule = (int) $product_detail['id_tax_rules_group'];
            $idOrderDetail = (int) $product_detail['id_order_detail'];
            $taxId = $this->getTaxId($orderTaxes, $idTaxRule, $idOrderDetail);
            $taxRateFromDetail = (float) $product_detail['tax_rate'];
            $taxRateFromId = $this->getTaxRateFromId($taxId);
            $taxRate = $taxRateFromDetail == $taxRateFromId ? $taxRateFromDetail : $taxRateFromId;
            $taxNameFromId = $this->getTaxNameFromId($taxId);

            if (Configuration::get('PS_CREAZIONE_FATTURA') == 1) {
                $natura = $this->getNaturaByIdTax($idTaxRule);
                if ((float)$taxRate == 0) {
                    $product->appendChild($domtree->createElement('FeVatNature', $natura));
                }
            }

            $product->appendChild($domtree->createElement('VatCode', $taxRate));
            //$taxName = Product::getTaxesInformations($product_detail)['tax_name'];
            $taxName = $product_detail['tax_calculator']->getTaxesName();
            //$VatDescription = !empty($taxName) ? $taxName : $this->getDefaultVatDescription($taxRate);
            $VatDescription = !empty($taxName) ? $taxName : $taxNameFromId;
            $product->appendChild($domtree->createElement('VatDescription', $VatDescription));

            if ($pdcSet) {
                $product->appendChild($domtree->createElement('IdPdc', $idPdc));
            }
        }
    }

    public function curlDownload($url, $data_string)
    {
        if (!function_exists('curl_init')) {
            $this->trace('curl is not installed');
            exit('Sorry, cURL is not installed!');
        }

        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);
        $config = [];
        $config['useragent'] = 'Mozilla/5.0 (Windows NT 6.2; WOW64; rv:17.0) Gecko/20100101 Firefox/17.0';
        curl_setopt($ch, CURLOPT_USERAGENT, $config['useragent']);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($ch, CURLOPT_POSTFIELDS, $data_string);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_HEADER, 0);
        curl_setopt($ch, CURLOPT_TIMEOUT, 60);
        if (!($output = curl_exec($ch))) {
            $this->trace('curl error', curl_getinfo($ch), curl_error($ch), curl_errno($ch));
        }
        curl_close($ch);

        return $output;
    }



    /**
     *  new function to get Carrier tax rates, for
     *  correct Natura handling
     */
    public function f24getCarriersTaxRate()
    {
        $context = Context::getContext();
        $id_lang = $context->language->id;
        $f24_carriers = Carrier::getCarriers($id_lang);
        $carrier_tax_rate = [];

        foreach ($f24_carriers as $val) {
            $id_carrier = $val['id_carrier'];
            if ((int) Tax::getCarrierTaxRate($id_carrier) == 0) {
                $carrier_tax_rate[$id_carrier] = Tax::getCarrierTaxRate($id_carrier);
            }
        }

        return $carrier_tax_rate;
    }

    /**
     *  Ottengo la descrizione dell'aliquota
     *  dall'id della tassa
     */
    public function getTaxNameFromId($idTax)
    {
        $context = Context::getContext();
        $id_lang = $context->language->id;
        $sql = 'SELECT * from ' . _DB_PREFIX_ .
               'tax_lang where id_tax = ' . (int) $idTax
               . ' AND id_lang =' . (int) $id_lang;
        $result = Db::getInstance()->getRow($sql);
        return $result['name'];
    }


    /**
     *  Ottengo l'aliquota dal'id della tassa
     */
    public function getTaxRateFromId($idTax)
    {
        $tax_rates = Tax::getTaxes();
        $taxRateResult = '';
        foreach ($tax_rates as $v) {
            if ((int) $v['id_tax'] == (int) $idTax) {
                $taxRateResult = $v['rate'];
            }
        }
        return (float) $taxRateResult;
    }

    /**
     *  Ottiene il taxId dal rate
     */
    public function getTaxIdFromRate($rate)
    {
        $tax_rates = Tax::getTaxes();
        $idFound = false;
        $idResult = '';

        foreach ($tax_rates as $v) {
            if ((number_format((float) $rate, 2, '.', '')) == (float) $v['rate']) {
                $idResult = $v['id_tax'];
            }
            $idFound = !empty($idResult);
            if ($idFound) {
                break;
            }
        }
        return $idResult;
    }


    // new function to get zero rates
    public function getZeroRates()
    {
        $tax_rates = Tax::getTaxes();
        $zero_rate = [];

        foreach ($tax_rates as $v) {
            if ((int) $v['rate'] == 0) {
                $id_tax = $v['id_tax'];
                $zero_rate[$id_tax] = $v['rate'];
            }
        }

        return $zero_rate;
    }

    /**
     * Visualizzo un messaggio di errore in caso di chiamata API fallita
     */
    public function getApiError($templates)
    {
        $apiErr = '';
        $failed_api_call = !empty(json_decode($templates, true));
        if ($failed_api_call) {
            $apiErr = $this->l('WARNING: connection to Fattura24 API failed, please contact our technical service');
        }

        if ($apiErr) {
            return $this->displayError($apiErr);
        }
    }

    public function getNaturaError()
    {
        $zeroRates = $this->getZeroRates(); // product rates
        $zeroShipping = $this->f24getCarriersTaxRate(); // shipping rates
        $fattEl = Configuration::get('PS_CREAZIONE_FATTURA') == 1; // solo per fatture elettroniche
        $emptyConfNatura = empty(Configuration::get('PS_CONF_NATURA_IVA'));
        $emptyConfSped = Configuration::get('PS_CONF_NATURA_SPEDIZIONE') == 'N0' && !empty($zeroShipping);
        $enteredConfNatura = preg_split('/-|;/', Configuration::get('PS_CONF_NATURA_IVA'));
        $selectedConfSped = Configuration::get('PS_CONF_NATURA_SPEDIZIONE');
        $oldNaturaArray = ['N2', 'N3', 'N6'];
        $errMsg = '';

        // different messages according to different scenario
        if (!empty($zeroRates) && $fattEl) {
            // primo blocco: campi vuoti
            if ($emptyConfNatura) {
                if ($emptyConfSped) {
                    $errMsg = $this->l('You should set up a Natura in both fields');
                } else {
                    $errMsg = $this->l('You should set up a Natura for each zero rate');
                }
            } elseif ($emptyConfSped) {
                $errMsg = $this->l('You should set up a Natura for each zero shipping rate');
            }

            // secondo blocco: campi compilati con valori non corretti
            if (!$emptyConfNatura) {
                $resultNatura = '';
                foreach ($enteredConfNatura as $naturaValue) {
                    if (array_search($naturaValue, $oldNaturaArray) !== false) {
                        $resultNatura = $naturaValue;
                    }
                }

                if (!empty($resultNatura)) {
                    $errMsg = $this->l('The code ' . $resultNatura . ' is no more allowed 
                    in invoices issued after 01/01/2021');
                }
            }
            if (in_array($selectedConfSped, $oldNaturaArray)) {
                $errMsg .= '<br />' . $this->l('The code ' . $selectedConfSped . ' in Shipping Natura is no more 
                allowed in invoices issued after 01/01/2021');
            }
        }

        if (!empty($errMsg)) {
            return $this->displayError($errMsg); // uso i metodi di prestashop
        } else {
            return '';
        }
    }

    protected function getParamSubstring($param)
    {
        $needlePos = strpos($param, '-');
        return (int) Tools::substr($param, 0, $needlePos);
    }

    public function getDefaultVatDescription($rate)
    {
        return 'IVA '.$rate . '%';
    }

    public function getNaturaVatDescription($idTaxRule)
    {
        if ($idTaxRule == 0) {
            $naturaSpedizione = Configuration::get('PS_CONF_NATURA_SPEDIZIONE');
            $lista = $this->getListaNatureNew();
            $needle = Tools::substr($naturaSpedizione, 0, 2);


            if ($naturaSpedizione == 'N0') {
                $naturaSpedizione = '';
                $VatDescription = 'IVA';
            } elseif (is_array($lista[$needle])) {
                $VatDescription = $lista[$needle][$naturaSpedizione];
            } else {
                $VatDescription = $lista[$naturaSpedizione];
            }
            return $VatDescription;
        }
        return false;
    }

    public function getNaturaByIdTax($idTax)
    {
        $output = Configuration::get('PS_CONF_NATURA_SPEDIZIONE');
        if ($idTax == '0') {
            // nuova gestione natura per prodotti con classe 'Nessuna tassa'
            $output = Configuration::get('PS_CONF_NATURA_SPEDIZIONE');

        // riporta la natura solo sui prodotti ad aliquota zero
        } elseif ($output == 'N0') {
            $naturaIvaStringConf = Configuration::get('PS_CONF_NATURA_IVA');
            $naturaIvaArrayConf = explode(';', $naturaIvaStringConf);
            foreach ($naturaIvaArrayConf as $oneNaturaConf) {
                if (strpos($oneNaturaConf, $idTax) !== false) {
                    $oneNaturaConfArray = explode('-', $oneNaturaConf);
                    $output = $oneNaturaConfArray[1];
                } else {
                    continue;
                }
            }
        }
        return $output;
    }
    /**
     * Nuova lista nature in conformità con nuove specifiche AdE
     */
    public function getListaNatureNew()
    {
        $tipiAliquoteIva = [
        'N0' => $this->l('Nessuna'),
        'N1' => ['N1' => 'N1 - Escluse ex art. 15'],
        'N2' => ['N2.1' => 'N2.1 - Non soggette ad IVA ai sensi degli artt. da 7 a 7-septies del DPR 633/72',
                 'N2.2' => 'N2.2 - Altri casi', ],
        'N3' => ['N3.1' => 'N3.1 - Esportazioni',
                 'N3.2' => 'N3.2 - Cessioni intracomunitarie',
                 'N3.3' => 'N3.3 - Cessioni verso San Marino',
                 'N3.4' => 'N3.4 - Operazioni assimilate alle cessioni all\'esportazione',
                 'N3.5' => 'N3.5 - A seguito di dichiarazioni d\'intento',
                 'N3.6' => 'N3.6 - Altre operazioni che non concorrono alla formazione del plafond', ],
        'N4' => ['N4' => 'N4 - Esenti'],
        'N5' => ['N5' => 'N5 - Regime del margine / IVA non esposta in fattura'],
        'N6' => ['N6.1' => 'N6.1 - Cessione di rottami e altri materiali di recupero',
                 'N6.2' => 'N6.2 - Cessione di oro e argento puro',
                 'N6.3' => 'N6.3 - Subappalto nel settore edile',
                 'N6.4' => 'N6.4 - Cessione di fabbricati',
                 'N6.5' => 'N6.5 - Cessione di telefoni cellulari',
                 'N6.6' => 'N6.6 - Cessione di prodotti elettronici',
                 'N6.7' => 'N6.7 - Prestazioni comparto edile e settori connessi',
                 'N6.8' => 'N6.8 - Operazioni settore energetico',
                 'N6.9' => 'N6.9 - Altri casi', ],
        'N7' => ['N7' => 'N7 - IVA assolta in altro stato UE'],
        ];

        return $tipiAliquoteIva;
    }

    // ottengo il timestamp attuale per il fuso orario di Roma
    public function now($fmt = 'Y-m-d H:i:s', $tz = 'Europe/Rome')
    {
        $timestamp = time();
        $dt = new \DateTime('now', new \DateTimeZone($tz));
        $dt->setTimestamp($timestamp);

        return $dt->format($fmt);
    }

    /**
     * Con questo metodo controllo se nel campo partita IVA
     * è stata immessa anche la sigla del paese di fatturazione
     * nei primi due caratteri: in quel caso la tronco
     */
    public function cleanVatNumber($billingCountry, $vatNumber)
    {
        $needle = Tools::substr(Tools::strtoupper($vatNumber), 0, 2);
        $result = $needle == $billingCountry ? Tools::substr($vatNumber, 2)
                  : $vatNumber;
        return $result;
    }

    public function getTaxId($orderTaxes, $idTaxRule, $idOrderDetail)
    {
        $taxId = 0;
        foreach ($orderTaxes as $tax) {
            if ($idTaxRule == (int)$tax['id_tax_rules_group']
               && $idOrderDetail == (int) $tax['id_order_detail']) {
                $taxId = (int) $tax['id_tax'];
            }
        }
        return $taxId;
    }

    // controllo se è stato pubblicato un aggiornamento del modulo
    public function checkNewVersion($urlCheckVersion)
    {
        try {
            $fileCheckVersion = fopen($urlCheckVersion, 'r');
            if ($fileCheckVersion) {
                $content = stream_get_contents($fileCheckVersion);
                fclose($fileCheckVersion);

                return $content;
            }
        } catch (Exception $e) {
            // non succede nulla
        }

        return false;
    }

    // se c'è un aggiornamento, visualizzo un messaggio
    public function getVersionCheckMessage()
    {
        //$versionCheckMessage = '';
        $urlCheckVersion = 'https://www.fattura24.com/prestashop/latest_version_1.7.txt';
        $checkNewVersion = $this->checkNewVersion($urlCheckVersion);
        if ($checkNewVersion) {
            $latest_version_date = Tools::substr($checkNewVersion, 0, 16);
            if ($latest_version_date !== $this->moduleDate) {
                //$urlDownload = Tools::substr($checkNewVersion, 17);
                $message = $this->l('A new version of the module has been released!');
                //$message .= $this->l(' Click <a href="' . $urlDownload . '" target="_blank">here</a> to download');
                return $this->displayWarning($message);
            }
        }

        return '';
    }

    public function getLogFileName()
    {
        $path = _PS_MODULE_DIR_ . 'fattura24/log/';
        $today = date('d_m_Y');
        $filename = $path . 'f24_trace_' . $today . '.log';
        if (!file_exists($filename)) {
            $this->delOldLogFiles();
        }

        return $filename;
    }

    public function delOldLogFiles()
    {
        $path = _PS_MODULE_DIR_ . 'fattura24/log/';
        $today = DateTime::createFromFormat('d_m_Y', date('d_m_Y'));
        foreach (glob($path . '*.log') as $file) {
            $creationDay = DateTime::createFromFormat('d_m_Y', date('d_m_Y', filectime($file)));
            $result = $creationDay->diff($today);
            $different_days = $result->d;
            if ($different_days > 30) {
                unlink($file);
            }
        }
    }

    // creo il file con cui traccio l'elaborato del modulo
    public function trace()
    {
        $logFileName = $this->getLogFileName();
        $fp = fopen($logFileName, 'a+');
        fprintf($fp, "%s: %s\n\n", $this->now(), var_export(func_get_args(), true));
        fclose($fp);
    }

    public function traceDebug($string)
    {
        $fp = fopen(_PS_MODULE_DIR_ . 'fattura24/log/traceDebug.log', 'a+');
        fprintf($fp, $string . "\n");
        fclose($fp);
    }

    // restituisce informazioni sull'ambiente (versione PHP, versione modulo etc. e le scrive nel file di log)
    public function getInfo()
    {
        return 'Versione del modulo di Fattura24: ' . $this->version . '\nVersione di Prestashop: '
            . _PS_VERSION_ . '\nVersione di PHP: ' . PHP_VERSION . '\nUrl negozio: ' . _PS_BASE_URL_;
    }
}
